# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the policy objects."""

import unittest

from pathlib import Path

import cvxpy as cp
import numpy as np
import pandas as pd

from cvxportfolio.policies import *
# from cvxportfolio.policies import SinglePeriodOptOLD, SinglePeriodOptNEW
from cvxportfolio.returns import *
from cvxportfolio.risks import *
from cvxportfolio.costs import *
from cvxportfolio.constraints import *
from cvxportfolio.errors import *


class TestPolicies(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        # cls.sigma = pd.read_csv(Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
        cls.returns = pd.read_csv(
            Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
        cls.volumes = pd.read_csv(
            Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
        cls.w_plus = cp.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
        cls.z = cp.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]

    def test_hold(self):
        hold = Hold()
        w = pd.Series(0.5, ["AAPL", "CASH"])
        self.assertTrue(np.all(
            hold._recursive_values_in_time(current_weights=w).values == np.zeros(2)))

    def test_rank_and_long_short(self):
        hold = Hold()
        w = pd.Series(0.25, ["AAPL", "TSLA", "GOOGL", "CASH"])
        signal = pd.Series([1, 2, 3], ["AAPL", "TSLA", "GOOGL"])
        num_long = 1
        num_short = 1
        target_leverage = 3.0
        rls = RankAndLongShort(
            signal=signal,
            num_long=num_long,
            num_short=num_short,
            target_leverage=target_leverage,
        )
        z = rls._recursive_values_in_time(t=None, current_weights=w)
        print(z)
        wplus = w + z
        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["TSLA"] == 0)
        self.assertTrue(wplus["AAPL"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)

        index = pd.date_range("2020-01-01", "2020-01-03")
        signal = pd.DataFrame(
            {
                "AAPL": pd.Series([1, 1.9, 3], index),
                "TSLA": pd.Series([3, 2.1, 1], index),
                "GOOGL": pd.Series([4, 4, 4], index),
            }
        )
        rls = RankAndLongShort(
            signal=signal,
            num_long=num_long,
            num_short=num_short,
            target_leverage=target_leverage,
        )
        z1 = rls._recursive_values_in_time(t=index[0], current_weights=w)
        print(z1)
        wplus = w + z1
        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["TSLA"] == 0)
        self.assertTrue(wplus["AAPL"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)
        z2 = rls._recursive_values_in_time(t=index[1], current_weights=w)
        print(z2)
        wplus = w + z2
        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["TSLA"] == 0)
        self.assertTrue(wplus["AAPL"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)
        z3 = rls._recursive_values_in_time(t=index[2], current_weights=w)
        wplus = w + z3
        self.assertTrue(wplus["CASH"] == 1)
        self.assertTrue(wplus["AAPL"] == 0)
        self.assertTrue(wplus["TSLA"] == -wplus["GOOGL"])
        self.assertTrue(np.abs(wplus[:-1]).sum() == 3)
        print(z3)

    def test_proportional_trade(self):

        a = pd.Series(1., self.returns.columns)
        a.iloc[-1] = 1 - sum(a.iloc[:-1])
        b = pd.Series(-1., self.returns.columns)
        b.iloc[-1] = 1 - sum(b.iloc[:-1])

        targets = pd.DataFrame({self.returns.index[3]: a,
                                self.returns.index[15]: b
                                }).T
        policy = ProportionalTradeToTargets(targets)

        policy._recursive_pre_evaluation(
            universe=self.returns.columns, backtest_times=self.returns.index)
        start_portfolio = pd.Series(
            np.random.randn(
                self.returns.shape[1]),
            self.returns.columns)
        start_portfolio.iloc[-1] = 1 - sum(start_portfolio.iloc[:-1])
        for t in self.returns.index[:17]:
            print(t)
            print(start_portfolio)

            trade = policy._recursive_values_in_time(
                t=t, current_weights=start_portfolio)
            start_portfolio += trade

            if t in targets.index:
                self.assertTrue(np.all(start_portfolio == targets.loc[t]))

        self.assertTrue(np.all(trade == 0.))

    def test_sell_all(self):
        start_portfolio = pd.Series(
            np.random.randn(
                self.returns.shape[1]),
            self.returns.columns)
        policy = SellAll()
        t = pd.Timestamp('2022-01-01')
        trade = policy._recursive_values_in_time(
            t=t, current_weights=start_portfolio)
        allcash = np.zeros(len(start_portfolio))
        allcash[-1] = 1
        assert isinstance(trade, pd.Series)
        assert np.allclose(allcash, start_portfolio + trade)

    def test_fixed_trade(self):
        fixed_trades = pd.DataFrame(
            np.random.randn(
                len(self.returns),
                self.returns.shape[1]),
            index=self.returns.index,
            columns=self.returns.columns)

        policy = FixedTrades(fixed_trades)
        t = self.returns.index[123]
        trade = policy._recursive_values_in_time(t=t, current_weights=pd.Series(
            0., self.returns.columns))
        self.assertTrue(np.all(trade == fixed_trades.loc[t]))

        t = pd.Timestamp('1900-01-01')
        trade = policy._recursive_values_in_time(t=t, current_weights=trade)
        self.assertTrue(np.all(trade == 0.))

    def test_fixed_weights(self):
        fixed_weights = pd.DataFrame(
            np.random.randn(
                len(self.returns),
                self.returns.shape[1]),
            index=self.returns.index,
            columns=self.returns.columns)

        policy = FixedWeights(fixed_weights)
        t = self.returns.index[123]
        trade = policy._recursive_values_in_time(t=t, current_weights=pd.Series(
            0., self.returns.columns))
        self.assertTrue(np.all(trade == fixed_weights.loc[t]))

        t = self.returns.index[111]
        trade = policy._recursive_values_in_time(
            t=t, current_weights=fixed_weights.iloc[110])
        self.assertTrue(np.allclose(
            trade + fixed_weights.iloc[110], fixed_weights.loc[t]))

        t = pd.Timestamp('1900-01-01')
        trade = policy._recursive_values_in_time(t=t, current_weights=trade)
        self.assertTrue(np.all(trade == 0.))

    def test_periodic_rebalance(self):

        target = pd.Series(np.random.uniform(
            size=self.returns.shape[1]), self.returns.columns)
        target /= sum(target)
        rebalancing_times = pd.date_range(start=self.returns.index[0], end=self.returns.index[-1],
                                          freq='7d')

        policy = PeriodicRebalance(target, rebalancing_times=rebalancing_times)
        init = pd.Series(np.random.randn(
            self.returns.shape[1]), self.returns.columns)

        trade = policy._recursive_values_in_time(
            t=rebalancing_times[0], current_weights=init)
        self.assertTrue(np.allclose(trade + init, target))

        trade = policy._recursive_values_in_time(t=rebalancing_times[0] + pd.Timedelta('1d'),
                                                 current_weights=init)
        self.assertTrue(np.allclose(trade, 0))

    def test_uniform(self):
        pol = Uniform()
        pol._recursive_pre_evaluation(self.returns.columns, self.returns.index)

        init = pd.Series(np.random.randn(
            self.returns.shape[1]), self.returns.columns)
        trade = pol._recursive_values_in_time(
            t=self.returns.index[123], current_weights=init)
        self.assertTrue(np.allclose((trade + init)[:-1],
                                    np.ones(self.returns.shape[1]-1)/(self.returns.shape[1]-1)))

    def test_proportional_rebalance(self):

        target = pd.Series(np.random.uniform(
            size=self.returns.shape[1]), self.returns.columns)
        target /= sum(target)
        target_matching_times = self.returns.index[::3]

        policy = ProportionalRebalance(
            target, target_matching_times=target_matching_times)
        policy._recursive_pre_evaluation(
            universe=self.returns.columns, backtest_times=self.returns.index)

        init = pd.Series(np.random.randn(
            self.returns.shape[1]), self.returns.columns)

        trade = policy._recursive_values_in_time(
            t=self.returns.index[1], current_weights=init)
        init += trade
        trade2 = policy._recursive_values_in_time(
            t=self.returns.index[2], current_weights=init)
        self.assertTrue(np.allclose(trade, trade2))
        self.assertTrue(np.allclose(trade + trade2 + init, target))

    def test_adaptive_rebalance(self):
        np.random.seed(0)
        target = pd.Series(
            np.random.uniform(
                size=self.returns.shape[1]),
            self.returns.columns)
        target /= sum(target)
        target = pd.DataFrame({ind: target for ind in self.returns.index}).T

        init = pd.Series(np.random.uniform(
            size=self.returns.shape[1]), self.returns.columns)
        init /= sum(init)

        for tracking_error in [0.01, .02, .05, .1]:
            policy = AdaptiveRebalance(target, tracking_error=tracking_error)
            trade = policy._recursive_values_in_time(
                t=self.returns.index[1], current_weights=init)
            self.assertTrue(np.allclose(init + trade, target.iloc[0]))

        for tracking_error in [.2, .5]:
            policy = AdaptiveRebalance(target, tracking_error=tracking_error)
            trade = policy._recursive_values_in_time(
                t=self.returns.index[1], current_weights=init)
            self.assertTrue(np.allclose(trade, 0.))

    def test_single_period_optimization(self):

        return_forecast = ReturnsForecast()
        risk_forecast = FullCovariance(kelly=False)
        tcost = TransactionCost(a=1E-3/2, pershare_cost=0., b=None, exponent=2)

        policy = SinglePeriodOptimization(
            return_forecast
            - 2 * risk_forecast
            # - TcostModel(half_spread=5 * 1E-4)  # , power=2)
            - tcost,
            constraints=[LongOnly(), LeverageLimit(1)],
            include_cash_return=False,
            # verbose=True,
            solver='ECOS')

        policy._recursive_pre_evaluation(
            universe=self.returns.columns, backtest_times=self.returns.index)
        policy._compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        result = policy._recursive_values_in_time(
            t=self.returns.index[121],
            current_weights=pd.Series(
                curw,
                self.returns.columns),
            current_portfolio_value=1000,
            past_returns=self.returns.iloc[:121],
            past_volumes=self.volumes.iloc[:121],
            current_prices=pd.Series(1., self.volumes.columns))

        cvxportfolio_result = pd.Series(result, self.returns.columns)

        print(cvxportfolio_result)

        # print(np.linalg.eigh(self.returns.iloc[:121, :-1].cov().values)[0])

        # REPLICATE WITH CVXPY

        # + np.outer(self.returns.iloc[:121, :-1].mean(), self.returns.iloc[:121, :-1].mean())
        COV = self.returns.iloc[:121, :-1].cov(ddof=0).values
        w = cp.Variable(self.N)
        cp.Problem(cp.Maximize(w[:-1].T @ self.returns.iloc[:121, :-1].mean().values -
                               2 * cp.quad_form(w[:-1], COV) -
                               5 * 1E-4 * cp.sum(cp.abs(w - curw)[:-1])
                               ),
                   [w >= 0, w <= 1, sum(w) == 1]
                   ).solve(solver='ECOS')

        cvxpy_result = pd.Series(w.value - curw, self.returns.columns)

        print(cvxpy_result)

        print(cvxportfolio_result - cvxpy_result)
        self.assertTrue(np.allclose(
            cvxportfolio_result - cvxpy_result, 0., atol=1e-5))

    def test_single_period_optimization_solve_twice(self):

        return_forecast = ReturnsForecast()
        risk_forecast = FullCovariance()

        policy = SinglePeriodOptimization(
            return_forecast
            - 2 * risk_forecast
            - TransactionCost(a=5 * 1E-4, pershare_cost=0., b=0.)  # , power=2)
            ,
            constraints=[LongOnly(), LeverageLimit(1)],
            # verbose=True,
            solver='ECOS')

        policy._recursive_pre_evaluation(
            universe=self.returns.columns, backtest_times=self.returns.index)
        policy._compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        result = policy._recursive_values_in_time(
            t=self.returns.index[134],
            current_weights=pd.Series(
                curw,
                self.returns.columns),
            current_portfolio_value=1000,
            past_returns=self.returns.iloc[:134],
            past_volumes=self.volumes.iloc[:134],
            current_prices=pd.Series(1., self.volumes.columns))

        self.assertFalse(np.allclose(result, 0.))

        cvxportfolio_result = pd.Series(result, self.returns.columns)

        curw += result

        result2 = policy._recursive_values_in_time(
            t=self.returns.index[134],
            current_weights=pd.Series(
                curw,
                self.returns.columns),
            current_portfolio_value=1000,
            past_returns=self.returns.iloc[:134],
            past_volumes=self.volumes.iloc[:134],
            current_prices=pd.Series(1., self.volumes.columns))

        self.assertTrue(np.allclose(result2, 0., atol=1e-7))

    def test_single_period_optimization_infeasible(self):

        return_forecast = ReturnsForecast()
        risk_forecast = FullCovariance()
        policy = SinglePeriodOptimization(
            return_forecast
            - 2 * risk_forecast
            - TransactionCost(a=5 * 1E-4, pershare_cost=0., b=0.)  # , power=2)
            ,
            constraints=[LongOnly(), LeverageLimit(1), MaxWeights(-1)],
            # verbose=True,
            solver='ECOS')

        policy._recursive_pre_evaluation(
            universe=self.returns.columns, backtest_times=self.returns.index)
        policy._compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        with self.assertRaises(PortfolioOptimizationError):
            result = policy._recursive_values_in_time(
                t=self.returns.index[134],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:134],
                past_volumes=self.volumes.iloc[:134],
                current_prices=pd.Series(1., self.volumes.columns))

    def test_single_period_optimization_unbounded(self):

        return_forecast = ReturnsForecast()
        risk_forecast = FullCovariance()
        policy = SinglePeriodOptimization(
            return_forecast            # - 2 * risk_forecast
            # - TransactionCost(spreads=10 * 1E-4, pershare_cost=0., b=0.)  # , power=2)
            ,
            constraints=[LongOnly(),  # LeverageLimit(1), MaxWeights(-1)
                         ],
            # verbose=True,
            solver='ECOS')

        policy._recursive_pre_evaluation(
            universe=self.returns.columns, backtest_times=self.returns.index)
        policy._compile_to_cvxpy()

        curw = np.zeros(self.N)
        curw[-1] = 1.

        with self.assertRaises(PortfolioOptimizationError):
            result = policy._recursive_values_in_time(
                t=self.returns.index[134],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:134],
                past_volumes=self.volumes.iloc[:134],
                current_prices=pd.Series(1., self.volumes.columns))

    def test_multi_period_optimization2(self):
        """Test that MPO1 and MPO2 and MPO5 return same if no tcost, and diff if tcost"""

        results = []
        for planning_horizon in [1, 2, 5]:
            return_forecast = ReturnsForecast()
            risk_forecast = FullCovariance()
            policy = MultiPeriodOptimization(
                return_forecast
                # - TcostModel(half_spread=5 * 1E-4)  # , power=2)
                - 10 * risk_forecast,
                constraints=[LongOnly(), LeverageLimit(1)],
                # verbose=True,
                planning_horizon=planning_horizon,
                solver='ECOS')

            policy._recursive_pre_evaluation(
                universe=self.returns.columns, backtest_times=self.returns.index)
            policy._compile_to_cvxpy()

            curw = np.zeros(self.N)
            curw[-1] = 1.

            results.append(policy._recursive_values_in_time(
                t=self.returns.index[67],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:67],
                past_volumes=self.volumes.iloc[:67],
                current_prices=pd.Series(1., self.volumes.columns).values))

        self.assertTrue(np.allclose(results[0], results[1], atol=1e-4))
        self.assertTrue(np.allclose(results[1], results[2], atol=1e-4))

        # with tcost

        results = []
        for planning_horizon in [1, 2, 5]:
            return_forecast = ReturnsForecast()
            risk_forecast = FullCovariance()
            policy = MultiPeriodOptimization(
                return_forecast
                - 10 * risk_forecast
                # - TcostModel(half_spread=5 * 1E-4)  # , power=2)
                - TransactionCost(a=25 * 1E-4, pershare_cost=0., b=0.),
                constraints=[LongOnly(), LeverageLimit(1)],
                # verbose=True,
                planning_horizon=planning_horizon,
                solver='ECOS')

            policy._recursive_pre_evaluation(
                universe=self.returns.columns, backtest_times=self.returns.index)
            policy._compile_to_cvxpy()

            curw = np.zeros(self.N)
            curw[-1] = 1.

            results.append(policy._recursive_values_in_time(
                t=self.returns.index[67],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:67],
                past_volumes=self.volumes.iloc[:67],
                current_prices=pd.Series(1., self.volumes.columns)))

        self.assertFalse(np.allclose(results[0], results[1], atol=1e-4))
        self.assertFalse(np.allclose(results[1], results[2], atol=1e-4))

    def test_multi_period_optimization_syntax(self):
        with self.assertRaises(SyntaxError):
            MultiPeriodOptimization([ReturnsForecast()], [])
        with self.assertRaises(SyntaxError):
            MultiPeriodOptimization([ReturnsForecast()], [[], []])
        with self.assertRaises(SyntaxError):
            MultiPeriodOptimization([ReturnsForecast()], None)
        with self.assertRaises(SyntaxError):
            MultiPeriodOptimization(ReturnsForecast())
        MultiPeriodOptimization(ReturnsForecast(), planning_horizon=1)

    def test_multi_period_optimization3(self):
        """Check that terminal constraint brings closer to benchmark."""

        np.random.seed(0)
        benchmark = np.random.uniform(size=self.returns.shape[1])
        benchmark /= sum(benchmark)
        benchmark = pd.Series(benchmark, self.returns.columns)

        diff_to_benchmarks = []
        for planning_horizon in [1, 2, 5]:

            return_forecast = ReturnsForecast()
            risk_forecast = FullCovariance()
            policy = MultiPeriodOptimization(
                return_forecast
                - 10 * risk_forecast
                # , power=2)
                - TransactionCost(a=5 * 1E-4, pershare_cost=0., b=0.),
                constraints=[LongOnly(), LeverageLimit(1)],
                # verbose=True,
                terminal_constraint=benchmark,
                planning_horizon=planning_horizon,
                solver='ECOS')

            policy._recursive_pre_evaluation(
                universe=self.returns.columns, backtest_times=self.returns.index)
            policy._compile_to_cvxpy()

            curw = np.zeros(self.N)
            curw[-1] = 1.

            diff_to_benchmarks.append(policy._recursive_values_in_time(
                t=self.returns.index[67],
                current_weights=pd.Series(
                    curw,
                    self.returns.columns),
                current_portfolio_value=1000,
                past_returns=self.returns.iloc[:67],
                past_volumes=self.volumes.iloc[:67],
                current_prices=pd.Series(1., self.volumes.columns)) + curw - benchmark)

        self.assertTrue(np.isclose(np.linalg.norm(diff_to_benchmarks[0]), 0.))
        self.assertTrue(np.linalg.norm(
            diff_to_benchmarks[0]) < np.linalg.norm(diff_to_benchmarks[1]))
        self.assertTrue(np.linalg.norm(
            diff_to_benchmarks[1]) < np.linalg.norm(diff_to_benchmarks[2]))


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines hyperparameter objects.
These are used currently as symbolic multipliers
of cost terms in Single and Multi Period Optimization policies
and can be iterated (and optimized over) automatically.
"""

import numpy as np
import copy
GAMMA_RISK_RANGE = [.5, 1., 2., 5., 10.]
GAMMA_COST_RANGE = [0., .1, .2, .5, 1., 2., 5., 10.]


__all__ = ['GammaRisk', 'GammaTrade', 'GammaHold']


class HyperParameter:
    """Base Hyper Parameter class.

    Implements arithmetic operations between hyper parameters.

    You can sum and multiply HPs between themselves and with scalars,
    and divide by a scalar. Arbitrary algebraic combination of these
    operations are supported.
    """

    def __mul__(self, other):
        if np.isscalar(other) or isinstance(other, HyperParameter):
            return CombinedHyperParameter([self], [other])
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        if np.isscalar(other):
            return CombinedHyperParameter([self], [1./other])
        return NotImplemented

    def __truediv__(self, other):
        return self.__div__(other)

    def __add__(self, other):
        if isinstance(other, HyperParameter):
            return CombinedHyperParameter([self, other], [1., 1.])
        return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        return self * (-1)

    def _collect_hyperparameters(self):
        return [self]


class CombinedHyperParameter(HyperParameter):
    """Algebraic combination of HyperParameters."""

    def __init__(self, left, right):
        self.left = left
        self.right = right

    @property
    def current_value(self):
        return sum([
            (le.current_value if hasattr(le, 'current_value') else le)
            * (ri.current_value if hasattr(ri, 'current_value') else ri)
            for le, ri in zip(self.left, self.right)])

    def _collect_hyperparameters(self):
        """Collect (not combined) hyperparameters."""
        result = []
        for el in self.left + self.right:
            if hasattr(el, '_collect_hyperparameters'):
                result += el._collect_hyperparameters()
        return result
        
    def __repr__(self):
        result = ''
        for le, ri in zip(self.left, self.right):
            result += str(le) + ' * ' + str(ri)
        return result


class RangeHyperParameter(HyperParameter):
    """Range Hyper Parameter.

    This is not meant to be used directly, look at
    its subclasses for ones that you can use.
    """

    def __init__(self, values_range, current_value):
        if not (current_value in values_range):
            raise SyntaxError('Initial value must be in the provided range')
        self.values_range = values_range
        self.current_value = current_value
        
    def __repr__(self):
        return self.__class__.__name__ \
            + f'(values_range={self.values_range}'\
            + f', current_value={self.current_value})'


class GammaRisk(RangeHyperParameter):
    """Multiplier of a risk term."""

    def __init__(self, values_range=GAMMA_RISK_RANGE, current_value=1.):
        super().__init__(values_range, current_value)


class GammaTrade(RangeHyperParameter):
    """Multiplier of a transaction cost term."""

    def __init__(self, values_range=GAMMA_COST_RANGE, current_value=1.):
        super().__init__(values_range, current_value)


class GammaHold(RangeHyperParameter):
    """Multiplier of a holding cost term."""

    def __init__(self, values_range=GAMMA_COST_RANGE, current_value=1.):
        super().__init__(values_range, current_value)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements realistic constraints to be used with
SinglePeriodOptimization and MultiPeriodOptimization policies,
or other Cvxpy-based policies.
"""


import cvxpy as cp
import numpy as np

from .estimator import CvxpyExpressionEstimator, DataEstimator
from .forecast import HistoricalFactorizedCovariance

__all__ = [
    "LongOnly",
    "LeverageLimit",
    "LongCash",
    "DollarNeutral",
    "ParticipationRateLimit",
    "MaxWeights",
    "MinWeights",
    "NoTrade",
    "FactorMaxLimit",
    "FactorMinLimit",
    "FixedFactorLoading",
    "MarketNeutral",
    "MinWeightsAtTimes",
    "MaxWeightsAtTimes",
    "TurnoverLimit",
    "MinCashBalance"
]


class BaseConstraint(CvxpyExpressionEstimator):
    """Base cvxpy constraint class."""


class BaseTradeConstraint(BaseConstraint):
    """Base class for constraints that operate on trades."""

    pass


class EqualityConstraint(BaseConstraint):
    """Base class for equality constraints.

    This class is not exposed to the user, each equality
    constraint inherits from this and overrides the 
    :func:`InequalityConstraint._compile_constr_to_cvxpy` and 
    :func:`InequalityConstraint._rhs` methods.

    We factor this code in order to streamline the
    design of :class:`SoftConstraint` costs.
    """

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile constraint to cvxpy."
        return self._compile_constr_to_cvxpy(w_plus, z, w_plus_minus_w_bm) == \
            self._rhs()

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Cvxpy expression of the left-hand side of the constraint.        
        """
        raise NotImplementedError

    def _rhs(self):
        """Cvxpy expression of the right-hand side of the constraint.        
        """
        raise NotImplementedError


class InequalityConstraint(BaseConstraint):
    """Base class for inequality constraints.

    This class is not exposed to the user, each inequality
    constraint inherits from this and overrides the 
    :func:`InequalityConstraint._compile_constr_to_cvxpy` and 
    :func:`InequalityConstraint._rhs` methods.

    We factor this code in order to streamline the
    design of :class:`SoftConstraint` costs.
    """

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile constraint to cvxpy."
        return self._compile_constr_to_cvxpy(w_plus, z, w_plus_minus_w_bm) <= \
            self._rhs()

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Cvxpy expression of the left-hand side of the constraint.        
        """
        raise NotImplementedError

    def _rhs(self):
        """Cvxpy expression of the right-hand side of the constraint.        
        """
        raise NotImplementedError


class CostInequalityConstraint(InequalityConstraint):
    """Linear inequality constraint applied to a cost term.

    The user does not interact with this class directly,
    it is returned by an expression such as `cost <= value`
    where `cost` is a :class:`BaseCost` instance and `value`
    is a scalar.
    """

    def __init__(self, cost, value):
        self.cost = cost
        self.value = DataEstimator(value, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile constraint to cvxpy."
        return self.cost._compile_to_cvxpy(w_plus, z, w_plus_minus_w_bm)

    def _rhs(self):
        return self.value.parameter

    def __repr__(self):
        return self.cost.__repr__() + ' <= ' + self.value.__repr__()


class BaseWeightConstraint(BaseConstraint):
    """Base class for constraints that operate on weights.

    Here we can implement a method to pass benchmark weights
    and make the constraint relative to it rather than to the null
    portfolio.
    """


class MarketNeutral(BaseWeightConstraint, EqualityConstraint):
    """Initial implementation of market neutrality.

    The benchmark portfolio weights are computed here
    (weighting by rolling averages of the market volumes)
    but instead should be their own class (used as well
    by risk models, ...).
    """

    def __init__(self):
        self.covarianceforecaster = HistoricalFactorizedCovariance()

    def _pre_evaluation(self, universe, backtest_times):
        self.market_vector = cp.Parameter(len(universe)-1)

    def _values_in_time(self, t, past_volumes, past_returns, **kwargs):
        tmp = past_volumes.iloc[-250:].mean()
        tmp /= sum(tmp)

        tmp2 = self.covarianceforecaster.current_value @ (
            self.covarianceforecaster.current_value.T @ tmp)
        # print(tmp2)
        self.market_vector.value = np.array(tmp2)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return w_plus[:-1].T @ self.market_vector

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return 0


class TurnoverLimit(BaseTradeConstraint, InequalityConstraint):
    """Turnover limit as a fraction of the portfolio value.

    See page 37 of the book.

    :param delta: constant or changing in time turnover limit
    :type delta: float or pd.Series
    """

    def __init__(self, delta):
        self.delta = DataEstimator(delta, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return .5 * cp.norm1(z[:-1])

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return self.delta.parameter


class ParticipationRateLimit(BaseTradeConstraint, InequalityConstraint):
    """A limit on maximum trades size as a fraction of market volumes.


    :param volumes: per-stock and per-day market volume estimates, or constant
        in time
    :type volumes: pd.Series or pd.DataFrame
    :param max_fraction_of_volumes: max fraction of market volumes that we're
        allowed to trade
    :type max_fraction_of_volumes: float, pd.Series, pd.DataFrame
    """

    def __init__(self, volumes, max_fraction_of_volumes=0.05):
        self.volumes = DataEstimator(volumes, compile_parameter=True)
        self.max_participation_rate = DataEstimator(
            max_fraction_of_volumes, compile_parameter=True)
        self.portfolio_value = cp.Parameter(nonneg=True)

    def _values_in_time(self, current_portfolio_value, **kwargs):
        self.portfolio_value.value = current_portfolio_value

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return cp.multiply(cp.abs(z[:-1]), self.portfolio_value)

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return cp.multiply(self.volumes.parameter, self.max_participation_rate.parameter)


class LongOnly(BaseWeightConstraint, InequalityConstraint):
    """A long only constraint.

    Imposes that at each point in time the post-trade
    weights are non-negative.

    """

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        return -w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return 0


class NoTrade(BaseTradeConstraint):
    """No-trade condition on name on periods(s)."""

    def __init__(self, asset, periods):
        self.asset = asset
        self.periods = periods

    def _pre_evaluation(self, universe, backtest_times):
        self.index = (universe.get_loc if hasattr(
            universe, 'get_loc') else universe.index)(self.asset)
        self.low = cp.Parameter()
        self.high = cp.Parameter()

    def _values_in_time(self, t, **kwargs):
        if t in self.periods:
            self.low.value = 0.
            self.high.value = 0.
        else:
            self.low.value = -100.
            self.high.value = +100.

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        return [z[self.index] >= self.low,
                z[self.index] <= self.high]


class LeverageLimit(BaseWeightConstraint, InequalityConstraint):
    """A limit on leverage.

    Leverage is defined as the :math:`\ell_1` norm of non-cash
    post-trade weights. Here we require that it is smaller than
    a given value.

    :param limit: constant or varying in time leverage limit
    :type limit: float or pd.Series
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return cp.norm(w_plus[:-1], 1)

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return self.limit.parameter


class MinCashBalance(BaseWeightConstraint):
    """Requires that the cash account is larger than c_min dollars.

    This uses logic to subtract cash used as margin for the short
    positions that is not documented in the book but is
    equivalent to the book definition's for long-only stock positions.
    """

    def __init__(self, c_min):
        self.c_min = DataEstimator(c_min)
        self.rhs = cp.Parameter()

    def _values_in_time(self, current_portfolio_value, **kwargs):
        self.rhs.value = self.c_min.current_value/current_portfolio_value

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Return a Cvxpy constraint."""
        # TODO clarify this
        realcash = (w_plus[-1] - 2 * cp.sum(cp.neg(w_plus[:-1])))
        return realcash >= self.rhs


class LongCash(MinCashBalance):
    """Requires that cash be non-negative."""

    def __init__(self):
        super().__init__(0.)


class DollarNeutral(BaseWeightConstraint, EqualityConstraint):
    """Long-short dollar neutral strategy."""

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return w_plus[-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return 1


class MaxWeights(BaseWeightConstraint, InequalityConstraint):
    """A max limit on weights.

    Attributes:
      limit: A series or number giving the weights limit.
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return self.limit.parameter


class MinWeights(BaseWeightConstraint, InequalityConstraint):
    """A min limit on weights.

    Attributes:
      limit: A series or number giving the weights limit.
    """

    def __init__(self, limit):
        self.limit = DataEstimator(limit, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return -w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return -self.limit.parameter


class MinMaxWeightsAtTimes(BaseWeightConstraint):

    def __init__(self, limit, times):
        self.base_limit = limit
        self.times = times

    def _pre_evaluation(self, universe, backtest_times):
        self.backtest_times = backtest_times
        self.limit = cp.Parameter()

    def _values_in_time(self, t, mpo_step, **kwargs):
        tidx = self.backtest_times.get_loc(t)
        nowtidx = tidx + mpo_step
        if (nowtidx < len(self.backtest_times)) and \
                (self.backtest_times[nowtidx] in self.times):
            self.limit.value = self.base_limit
        else:
            self.limit.value = 100 * self.sign


class MinWeightsAtTimes(MinMaxWeightsAtTimes, InequalityConstraint):

    sign = -1.  # used in _values_in_time of parent class

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return -w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return -self.limit


class MaxWeightsAtTimes(MinMaxWeightsAtTimes, InequalityConstraint):

    sign = 1.  # used in _values_in_time of parent class

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return self.limit


class FactorMaxLimit(BaseWeightConstraint, InequalityConstraint):
    """A max limit on portfolio-wide factor (e.g. beta) exposure.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time 
        and has the assets names as columns, and it represents factor 
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex: 
        first level is time, second level are assets' names (if changing 
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each 
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param limit: Factor limit, either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type limit: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(
            factor_exposure, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True, 
            ignore_shape_check=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return self.factor_exposure.parameter.T @ w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return self.limit.parameter


class FactorMinLimit(BaseWeightConstraint, InequalityConstraint):
    """A min limit on portfolio-wide factor (e.g. beta) exposure.

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time 
        and has the assets names as columns, and it represents factor 
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex: 
        first level is time, second level are assets' names (if changing 
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each 
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param limit: Factor limit, either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type limit: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, limit):
        self.factor_exposure = DataEstimator(
            factor_exposure, compile_parameter=True)
        self.limit = DataEstimator(limit, compile_parameter=True,
            ignore_shape_check=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return -self.factor_exposure.parameter.T @ w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return -self.limit.parameter


class FixedFactorLoading(BaseWeightConstraint, EqualityConstraint):
    """A constraint to fix portfolio loadings to a set of factors.

    This can be used to impose market neutrality, 
    a certain portfolio-wide alpha, ....

    :param factor_exposure: Series or DataFrame giving the factor exposure.
        If Series it is indexed by assets' names and represents factor
        exposures constant in time. If DataFrame it is indexed by time 
        and has the assets names as columns, and it represents factor 
        exposures that change in time. In the latter case an observation
        must be present for every point in time of a backtest.
        If you want you can also pass multiple factor exposures at once:
        as a dataframe indexed by assets' names and whose columns are the
        factors (if constant in time), or a dataframe with multiindex: 
        first level is time, second level are assets' names (if changing 
        in time). However this latter usecase is probably better served
        by making multiple instances of this constraint, one for each 
        factor.
    :type factor_exposure: pd.Series or pd.DataFrame
    :param target: Target portfolio factor exposures, 
        either constant or varying in time. Use
        a DataFrame if you pass multiple factors as once.
    :type target: float or pd.Series or pd.DataFrame
    """

    def __init__(self, factor_exposure, target):
        self.factor_exposure = DataEstimator(
            factor_exposure, compile_parameter=True)
        self.target = DataEstimator(target, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        "Compile left hand side of the constraint expression."
        return self.factor_exposure.parameter.T @ w_plus[:-1]

    def _rhs(self):
        "Compile right hand side of the constraint expression."
        return self.target.parameter



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the constraints objects."""

import unittest
from pathlib import Path

import numpy as np
import pandas as pd
import cvxpy as cp
import cvxportfolio as cvx


def listvalue(li):
    return all([el.value() for el in li])


class TestConstraints(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        cls.sigma = pd.read_csv(
            Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
        cls.returns = pd.read_csv(
            Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
        cls.volumes = pd.read_csv(
            Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
        cls.w_plus = cp.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
        cls.z = cp.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]

    def build_constraint(self, constraint, t=None):
        """Initialize constraint, build expression, and point it to given time."""
        constraint._recursive_pre_evaluation(
            self.returns.columns, self.returns.index)
        cvxpy_expression = constraint._compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        constraint._recursive_values_in_time(t=pd.Timestamp(
            "2020-01-01") if t is None else t, current_portfolio_value=1000)
        return cvxpy_expression

    def test_long_only(self):
        model = cvx.LongOnly()
        cons = self.build_constraint(model)
        self.w_plus.value = np.ones(self.N)
        self.assertTrue(cons.value())
        # self.assertTrue(listvalue(cons))
        self.w_plus.value = -np.ones(self.N)
        self.assertFalse(cons.value())
        # self.assertFalse(listvalue(cons))
        # model = cvx.LongOnly(nocash=True)
        # cons = self.build_constraint(model)
        # self.w_plus.value = np.ones(self.N)
        # self.assertFalse(listvalue(cons))

    def test_long_cash(self):
        model = cvx.LongCash()
        cons = self.build_constraint(model)
        self.w_plus.value = np.ones(self.N)
        self.assertTrue(cons.value())
        tmp = np.ones(self.N)
        tmp[-1] = -1
        self.w_plus.value = tmp
        self.assertTrue(cons.is_dcp())
        self.assertFalse(cons.value())

    def test_min_cash(self):
        model = cvx.MinCashBalance(10000)  # USD
        cons = self.build_constraint(model)
        self.w_plus.value = np.zeros(self.N)
        self.w_plus.value[-1] = 1
        model._recursive_values_in_time(t=pd.Timestamp(
            "2020-01-01"), current_portfolio_value=10001)
        self.assertTrue(cons.value())
        model._recursive_values_in_time(t=pd.Timestamp(
            "2020-01-01"), current_portfolio_value=9999)
        self.assertFalse(cons.value())

    def test_dollar_neutral(self):
        model = cvx.DollarNeutral()
        cons = self.build_constraint(model)
        tmpvalue = np.zeros(self.N)
        tmpvalue[-1] = 1 - sum(tmpvalue[:-1])
        self.w_plus.value = tmpvalue
        self.assertTrue(cons.value())
        tmpvalue = np.ones(self.N)
        tmpvalue[-1] = 1 - sum(tmpvalue[:-1])
        self.w_plus.value = tmpvalue
        self.assertFalse(cons.value())

    def test_leverage_limit(self):
        model = cvx.LeverageLimit(2)
        cons = self.build_constraint(model)
        self.w_plus.value = np.ones(self.N) / self.N
        self.assertTrue(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())
        model = cvx.LeverageLimit(7)
        cons = self.build_constraint(model)
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_leverage_limit_in_time(self):
        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = 7
        model = cvx.LeverageLimit(limits)
        cons = self.build_constraint(model, t=self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())
        model._recursive_values_in_time(t=self.returns.index[2])
        self.assertFalse(cons.value())

    def test_max_weights(self):
        model = cvx.MaxWeights(2)
        cons = self.build_constraint(model)
        self.w_plus.value = np.ones(self.N) / self.N
        self.assertTrue(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())

        model = cvx.MaxWeights(7)
        cons = self.build_constraint(model)

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = 7

        model = cvx.MaxWeights(limits)
        cons = self.build_constraint(model, t=self.returns.index[1])

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())
        model._recursive_values_in_time(t=self.returns.index[2])
        self.assertFalse(cons.value())

    def test_min_weights(self):
        model = cvx.MinWeights(2)
        cons = self.build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())
        model = cvx.MinWeights(-3)
        cons = self.build_constraint(model, self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

        limits = pd.Series(index=self.returns.index, data=2)
        limits.iloc[1] = -3
        model = cvx.MinWeights(limits)
        cons = self.build_constraint(model, t=self.returns.index[1])
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[-1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())
        model._recursive_values_in_time(t=self.returns.index[2])
        self.assertFalse(cons.value())

    def test_factor_max_limit(self):

        model = cvx.FactorMaxLimit(
            np.ones((self.N - 1, 2)), np.array([0.5, 1]))
        cons = self.build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertFalse(cons.value())

        model = cvx.FactorMaxLimit(np.ones((self.N - 1, 2)), np.array([4, 4]))
        cons = self.build_constraint(model, self.returns.index[1])

        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_factor_min_limit(self):

        model = cvx.FactorMinLimit(
            np.ones((self.N - 1, 2)), np.array([0.5, 1]))
        cons = self.build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3

        self.w_plus.value = tmp
        self.assertTrue(cons.value())

        model = cvx.FactorMinLimit(np.ones((self.N - 1, 2)), np.array([4, 4]))
        cons = self.build_constraint(model, self.returns.index[1])
        # cons = model.weight_expr(t, self.w_plus, None, None)[0]
        tmp = np.zeros(self.N)
        tmp[0] = 4
        # tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_fixed_alpha(self):
        model = cvx.FixedFactorLoading(np.ones((self.N - 1, 1)), 1)
        cons = self.build_constraint(model, self.returns.index[1])

        self.w_plus.value = np.ones(self.N) / self.N
        self.assertFalse(cons.value())
        tmp = np.zeros(self.N)
        tmp[0] = 4
        tmp[1] = -3
        self.w_plus.value = tmp
        self.assertTrue(cons.value())

    def test_turnover_limit(self):
        model = cvx.TurnoverLimit(0.1)
        cons = self.build_constraint(model)
        self.z.value = np.zeros(self.N)
        self.z.value[-1] = -sum(self.z.value[:-1])
        self.assertTrue(cons.value())

        self.z.value[1] = 0.2
        self.z.value[-1] = -sum(self.z.value[:-1])
        self.assertTrue(cons.value())

        self.z.value[2] = -0.01
        self.z.value[-1] = -sum(self.z.value[:-1])
        self.assertFalse(cons.value())

    def test_participation_rate(self):
        """Test trading constraints."""

        t = self.returns.index[1]

        # avg daily value limits.
        value = 1e6
        model = cvx.ParticipationRateLimit(
            self.volumes, max_fraction_of_volumes=0.1)
        model._recursive_pre_evaluation(
            self.returns.columns, self.returns.index)
        cons = model._compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        model._recursive_values_in_time(t=t, current_portfolio_value=value)
        print(model.portfolio_value.value)
        # cons = model.weight_expr(t, None, z, value)[0]
        tmp = np.zeros(self.N)
        tmp[:-1] = self.volumes.loc[t].values / value * 0.05
        self.z.value = tmp
        self.assertTrue(cons.value())
        self.z.value = -100 * self.z.value  # -100*np.ones(n)
        self.assertFalse(cons.value())


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['DataError', 'MissingValuesError', 'ForeCastError',
           'PortfolioOptimizationError', 'Bankruptcy', 'ConvexSpecificationError',
           'ConvexityError']


class DataError(Exception):
    """Base class for exception related to data."""

    pass


class MissingValuesError(DataError):
    """Cvxportfolio tried to access numpy.nan values."""

    pass


class ForeCastError(DataError):
    """Forecast procedure failed."""
    pass


class PortfolioOptimizationError(Exception):
    """Errors with portfolio optimization problems."""
    pass


class Bankruptcy(Exception):
    """A backtest resulted in a bankruptcy."""
    pass


class ConvexSpecificationError(SyntaxError):
    def __init__(self, term):
        super().__init__(
            "The convex optimization term"
            f" {term} does not follow the convex optimization specifications."
            " This could be due to a mis-specified custom term, or a non-convex"
            " cost inequality used as constraint (e.g., `[cvx.FullCovariance() >= constant]`)."
        )


class ConvexityError(SyntaxError):
    def __init__(self, cost):
        super().__init__(
            f"The cost term {cost}"
            " is not concave. (We need concavity since we maximize it.) You probably have"
            " a cost's multiplier with the wrong sign.")



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements cost functions used by optimization-based policies.

Currently these are two: :class:`StocksTransactionCost` and :class:`StocksHoldingCost`.

The default parameters are chosen to approximate real costs for the stock market
as well as possible. 
"""

import cvxpy as cp
import numpy as np
import pandas as pd
import copy
import inspect

from .estimator import CvxpyExpressionEstimator,  DataEstimator
from .utils import periods_per_year
from .hyperparameters import HyperParameter
from .constraints import EqualityConstraint, InequalityConstraint, CostInequalityConstraint
from .errors import *

__all__ = ["HoldingCost", "TransactionCost", "SoftConstraint",
           "StocksTransactionCost", "StocksHoldingCost"]


class BaseCost(CvxpyExpressionEstimator):
    """Base class for cost objects (and also risks).

    Here there is some logic used to implement the algebraic operations.
    See also :class:`CombinedCost`.
    """

    def _simulate(self, *args, **kwargs):
        """Simulate cost, used by market simulator.

        Look at its invocation in ``MarketSimulator`` for its list of 
        arguments.

        Cost classes that are meant to be used in the simulator
        should implement this.
        """
        raise NotImplementedError

    def __mul__(self, other):
        """Multiply by constant."""
        if not (np.isscalar(other) or isinstance(other, HyperParameter)):
            raise SyntaxError("You can only multiply cost by a scalar "
                              + "or a HyperParameter instance. (Have you instantiated it?)")
        return CombinedCosts([self], [other])

    def __add__(self, other):
        """Add cost expression to another cost expression.

        Idea is to create a new CombinedCost class that
        implements `_compile_to_cvxpy` and _recursive_values_in_time
        by summing over costs.

        """
        if isinstance(other, CombinedCosts):
            return other + self
        return CombinedCosts([self, other], [1.0, 1.0])

    def __rmul__(self, other):
        """Multiply by constant."""
        return self * other

    def __radd__(self, other):
        """Add cost expression to another cost."""
        return self + other

    def __neg__(self):
        """Take negative of cost."""
        return self * -1

    def __sub__(self, other):
        """Subtract other cost."""
        return self.__add__(-other)

    def __rsub__(self, other):
        """Subtract from other cost."""
        return other.__add__(-self)

    def __le__(self, other):
        """self <= other, return CostInequalityConstraint.

        For now we check here the type of "other"
        but it would be nicer to have CostInequalityConstraint's
        internal DataEstimator throw NotImplemented instead.
        """
        if isinstance(other, float) \
                or isinstance(other, int) \
                or isinstance(other, pd.Series):
            return CostInequalityConstraint(self, other)
        return NotImplemented

    def __ge__(self, other):
        """self >= other, return CostInequalityConstraint."""
        return (-self).__le__(other)


class CombinedCosts(BaseCost):
    """Algebraic combination of :class:`BaseCost` instances.

    :param costs: instances of :class:`BaseCost`
    :type costs: list 
    :param multipliers: floats that multiply the ``costs``
    :type multipliers: list
    """

    def __init__(self, costs, multipliers):
        for cost in costs:
            if not isinstance(cost, BaseCost):
                raise SyntaxError(
                    "You can only sum cost instances to other cost instances.")
        self.costs = costs
        self.multipliers = multipliers

    def __add__(self, other):
        """Add other (combined) cost to self."""
        if isinstance(other, CombinedCosts):
            return CombinedCosts(self.costs + other.costs,
                                 self.multipliers + other.multipliers)
        else:
            return CombinedCosts(self.costs + [other], self.multipliers + [1.0])

    def __mul__(self, other):
        """Multiply by constant."""
        return CombinedCosts(self.costs,
                             [el * other for el in self.multipliers])

    def _recursive_pre_evaluation(self, *args, **kwargs):
        """Iterate over constituent costs."""
        [el._recursive_pre_evaluation(*args, **kwargs) for el in self.costs]

    def _recursive_values_in_time(self, **kwargs):
        """Iterate over constituent costs."""
        [el._recursive_values_in_time(**kwargs) for el in self.costs]

    def _compile_to_cvxpy(self, w_plus, z, portfolio_value):
        """Iterate over constituent costs."""
        self.expression = 0
        for multiplier, cost in zip(self.multipliers, self.costs):
            add = (multiplier.current_value 
                if hasattr(multiplier, 'current_value') else multiplier) * \
                    cost._compile_to_cvxpy(w_plus, z, portfolio_value)
            if not add.is_dcp():
                raise ConvexSpecificationError(cost * multiplier)
            if not add.is_concave():
                raise ConvexityError(cost * multiplier)
            self.expression += add
        return self.expression

    def _collect_hyperparameters(self):
        return sum([el._collect_hyperparameters() for el in self.costs], []) + \
            sum([el._collect_hyperparameters() for el in self.multipliers if
                hasattr(el, '_collect_hyperparameters')], [])

    def __repr__(self):
        """Pretty-print."""
        result = ''
        for i, (mult, cost) in enumerate(zip(self.multipliers, self.costs)):
            if not isinstance(mult, HyperParameter):
                if mult == 0:
                    continue
                if mult < 0:
                    result += ' - ' if i > 0 else '-'
                else:
                    result += ' + ' if i > 0 else ''
                    result += (str(abs(mult)) + ' * ' if abs(mult) != 1 else '')
            else:
                result += str(mult) + ' * '
            result += cost.__repr__()
        return result


class SoftConstraint(BaseCost):
    """Soft constraint cost.

    :param constraint: cvxportfolio constraint instance whose
        violation we penalize
    :type constraint: cvx.BaseConstraint
    """

    def __init__(self, constraint):
        self.constraint = constraint

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost to cvxpy expression."""
        try:
            expr = (self.constraint._compile_constr_to_cvxpy(w_plus, z, w_plus_minus_w_bm)
                    - self.constraint._rhs())
        except AttributeError:
            raise SyntaxError(f"{self.__class__.__name__} can only be used with"
                              " EqualityConstraint or InequalityConstraint instances.")
        if isinstance(self.constraint, EqualityConstraint):
            return cp.sum(cp.abs(expr))
        if isinstance(self.constraint, InequalityConstraint):
            return cp.sum(cp.pos(expr))


def _annual_percent_to_per_period(value, ppy):
    """Transform annual percent to per-period return."""
    return np.exp(np.log(1 + value / 100) / ppy) - 1


class HoldingCost(BaseCost):
    r"""Generic holding cost model, as described in page 11 of the book.
    
    There are two ways to use this class. Either in the costs attribute
    of a :class:`MarketSimulator`, in which case the costs are evaluated
    on the post-trade dollar positions :math:`h^+_t`. Or, 
    as part of the objective function (or as a constraint!) 
    of a :class:`SinglePeriodOptimization` 
    or :class:`MultiPeriodOptimization` trading policy, in which case they
    are evaluated on the post-trade weights :math:`w_t + z_t`. The mathematical
    form is the same (see the discussion at pages 11-12 of the book).
    
    This particular implementation represents the following objective terms
    (expressed here in terms of the post-trade dollar positions):
    
    .. math::
 
        s^T_t {(h^+_t)}_- + l^T_t {(h^+_t)}_+ - d^T_t h^+_t
    
    where :math:`s_t` are the (short) borrowing fees, 
    :math:`l_t` are the fees on long positions,
    and :math:`d_t` are dividend rates (their sign is flipped because
    the costs are deducted from the cash account at each period). See
    below for their precise definition.
    
    Example usage as simulator cost:
    
    .. code-block:: python
    
        borrow_fees = pd.Series([5, 10], index=['AAPL', 'ZM'])
        simulator = cvx.MarketSimulator(['AAPL', 'ZM'], 
            costs=cvx.HoldingCost(short_fees=borrow_fees))
    
    Example usage as trading policy cost:
    
    .. code-block:: python
        
        objective = cvx.ReturnsForecast() - 5 * cvx.FullCovariance() \
            - cvx.HoldingCost(short_fees=10)
        constraints = [cvx.LeverageLimit(3)]    
        policy = cvx.SinglePeriodOptimization(objective, constraints)
    
    :param short_fees: Short borrowing fees expressed as annual percentage;
        you can provide them as a float (constant for all times and all 
        assets), a :class:`pd.Series` indexed by time (constant for all
        assets but varying in time) or by assets' names (constant in time
        but varying across assets), or a :class:`pd.DataFrame` indexed by 
        time and whose columns are the assets' names, if varying both
        in time and across assets. If you use a time-indexed pandas object
        be careful to include all times at which a backtest is evaluated
        (otherwise you'll get a :class:`MissingValueError` exception). If `None`,
        the term is ignored.
    :type short_fees: float, pd.Series, pd.DataFrame or None
    :param long_fees: Fees on long positions expressed as annual percentage;
        same convention as above applies.
    :type long_fees: float, pd.Series, pd.DataFrame or None
    :param dividends: Dividend rates per period. Dividends are already included in the market
        returns by the default data interface (based on Yahoo Finance "adjusted prices")
        and thus this parameter should not be used in normal circumstances.
    :type dividends: float, pd.Series, pd.DataFrame or None
    :param periods_per_year: How many trading periods are there in a year, for
        example 252 (for trading days in the US). This is
        only relevant when using this class as part of a trading policy. If you leave
        this to `None` the following happens.
        The value of periods per year are estimated at each period by looking at 
        the past market returns at that point in time: the number of past periods is divided by 
        the timestamp of the most recent period minus the timestamp of the
        first period (in years). That works well in most cases where there is enough history
        (say, a few years) and saves the user from having to manually enter this.
        If instead you use this object as a cost in a market simulator the parameter has no
        effect. (The length of each trading period in the simulation is known and so the per-period
        rates are evaluated exactly. For example, the rate over a weekend will be higher
        than overnight.)
    :type periods_per_year: float or None
    """

    def __init__(self, short_fees=None, long_fees=None, dividends=None,
                 periods_per_year=None):

        self.short_fees = None if short_fees is None else DataEstimator(short_fees)
        self.long_fees = None if long_fees is None else DataEstimator(long_fees)
        self.dividends = None if dividends is None else DataEstimator(dividends)
        self.periods_per_year = periods_per_year

    def _pre_evaluation(self, universe, backtest_times):
        """Initialize cvxpy parameters.
        
        We don't use the parameter from DataEstimator because we need to
        divide the value by periods_per_year.
        """

        if not (self.short_fees is None):
            self._short_fees_parameter = cp.Parameter(len(universe) - 1, nonneg=True)

        if not (self.long_fees is None):
            self._long_fees_parameter = cp.Parameter(len(universe) - 1, nonneg=True)

        if not (self.dividends is None):
            self._dividends_parameter = cp.Parameter(len(universe) - 1)

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update cvxpy parameters.
        
        We compute the estimate of periods per year from past returns
        (if not provided by the user) and populate the cvxpy parameters
        with the current values of the user-provided data, transformed
        to per-period.
        """
        
        if not ((self.short_fees is None) 
                and (self.long_fees is None) 
                and (self.dividends is None)):
            ppy = periods_per_year(past_returns.index) if self.periods_per_year is None else \
                self.periods_per_year

        if not (self.short_fees is None):
            self._short_fees_parameter.value = \
                np.ones(past_returns.shape[1]-1) * \
                     _annual_percent_to_per_period(self.short_fees.current_value, ppy)

        if not (self.long_fees is None):
            self._long_fees_parameter.value = \
                np.ones(past_returns.shape[1]-1) * \
                     _annual_percent_to_per_period(self.long_fees.current_value, ppy)

        if not (self.dividends is None):
            self._dividends_parameter.value = \
                np.ones(past_returns.shape[1]-1) * self.dividends.current_value


    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost to cvxpy expression."""

        expression = 0.
        
        if not (self.short_fees is None):
            expression += self._short_fees_parameter.T @ cp.neg(w_plus[:-1])
        
        if not (self.long_fees is None):
            expression += self._long_fees_parameter.T @ cp.pos(w_plus[:-1])

        if not (self.dividends is None):
            # we have a minus sign because costs are deducted from PnL
            expression -= self._dividends_parameter.T @ w_plus[:-1]
            
        assert expression.is_convex()
        
        return expression

    
    def _simulate(self, t, h_plus, current_and_past_returns, t_next, **kwargs):
        """
        TODO: make sure simulator cost sign convention is the same as optimization cost! OK
        TODO: make sure DataEstimator returns np.array of correct size! ~OK
        TODO: make sure simulator cost estimators are recursively evaluated! ~OK
        """
        
        year_divided_by_period = pd.Timedelta('365.24d') / (t_next - t)

        cost = 0. 
        
        # TODO this is a temporary fix, we should plug this into a recursive tree
        for est in [self.short_fees, self.long_fees, self.dividends]:
            if not (est is None):
                est._recursive_pre_evaluation(universe=h_plus.index, backtest_times=[t])
                est._recursive_values_in_time(t=t)

        if not (self.short_fees is None):
            cost += np.sum(_annual_percent_to_per_period(
                self.short_fees.current_value, year_divided_by_period) * (-np.minimum(h_plus[:-1], 0.)))
        
        if not (self.long_fees is None):
            cost += np.sum(_annual_percent_to_per_period(
                self.long_fees.current_value, year_divided_by_period) * np.maximum(h_plus[:-1], 0.))

        if not (self.dividends is None):
            # we have a minus sign because costs are deducted from PnL
            cost -= np.sum(self.dividends.current_value * h_plus[:-1])

        return cost


class StocksHoldingCost(HoldingCost):
    """A simplified version of :class:`HoldingCost` for the holding cost of stocks.
    
    This implements the simple model describe at page 11 of the book, *i.e.*
    the cost (in terms of the post-trade dollar positions):
    
    .. math::
 
        s^T_t {(h^+_t)}_-
    
    This class is a specialized version of :class:`HoldingCost`, and you should
    read its documentation for all details. Here we
    drop most of the parameters and use the default values explained above.
    We use a default value of :math:`5\%` annualized borrowing
    fee which is a rough (optimistic) approximation of the cost
    of shorting liquid US stocks. This cost is included **by default**
    in :class:`StockMarketSimulator`, the market simulator specialized
    to US (liquid) stocks.

    :param short_fees: Same as in :class:`HoldingCost`.
    :type short_fees: float, pd.Series, pd.DataFrame or None
    """

    def __init__(self, short_fees=5):

        super().__init__(short_fees=short_fees)


class TransactionCost(BaseCost):
    """This is a generic model for transaction cost of financial assets.

    Currently it is not meant to be used directly. Look at
    :class:`StocksTransactionCost` for its version specialized
    to the stock market.
    """

    def __init__(self, a=None, pershare_cost=None, b=0., window_sigma_est=None,
                 window_volume_est=None, exponent=None):

        self.a = None if a is None else DataEstimator(a)
        self.pershare_cost = None if pershare_cost is None else DataEstimator(
            pershare_cost)
        self.b = None if b is None else DataEstimator(b)
        self.window_sigma_est = window_sigma_est
        self.window_volume_est = window_volume_est
        self.exponent = exponent

    def _pre_evaluation(self, universe, backtest_times):
        if self.a is not None or self.pershare_cost is not None:
            self.first_term_multiplier = cp.Parameter(
                len(universe)-1, nonneg=True)
        if self.b is not None:
            self.second_term_multiplier = cp.Parameter(
                len(universe)-1, nonneg=True)

    def _values_in_time(self, t,  current_portfolio_value, past_returns,
                        past_volumes, current_prices, **kwargs):

        tmp = 0.

        if self.a is not None:
            _ = self.a.current_value
            tmp += _ * \
                np.ones(past_returns.shape[1]-1) if np.isscalar(_) else _
        if self.pershare_cost is not None:
            if current_prices is None:
                raise SyntaxError(
                    "If you don't provide prices you should set pershare_cost to None")
            tmp += self.pershare_cost.current_value / current_prices.values

        if self.a is not None or self.pershare_cost is not None:
            self.first_term_multiplier.value = tmp

        if self.b is not None:

            if (self.window_sigma_est is None) or \
                    (self.window_volume_est is None):
                ppy = periods_per_year(past_returns.index)
            windowsigma = ppy if (
                self.window_sigma_est is None) else self.window_sigma_est
            windowvolume = ppy if (
                self.window_volume_est is None) else self.window_volume_est

            # TODO refactor this with forecast.py logic
            sigma_est = np.sqrt(
                (past_returns.iloc[-windowsigma:, :-1]**2).mean()).values
            volume_est = past_volumes.iloc[-windowvolume:].mean().values

            self.second_term_multiplier.value = self.b.current_value * sigma_est * \
                (current_portfolio_value /
                 volume_est) ** ((2 if self.exponent is None else self.exponent) - 1)

    def _simulate(self, t, u, current_and_past_returns,
                  current_and_past_volumes, current_prices, **kwargs):

        if self.window_sigma_est is None:
            windowsigma = periods_per_year(current_and_past_returns.index)
        else:
            windowsigma = self.window_sigma_est

        exponent = (1.5 if self.exponent is None else self.exponent)

        sigma = np.std(
            current_and_past_returns.iloc[-windowsigma:, :-1], axis=0)

        result = 0.
        if self.pershare_cost is not None:
            if current_prices is None:
                raise SyntaxError(
                    "If you don't provide prices you should set pershare_cost to None")
            result += self.pershare_cost._recursive_values_in_time(t) * int(
                sum(np.abs(u.iloc[:-1] + 1E-6) / current_prices.values))

        if self.a is not None:
            result += sum(self.a._recursive_values_in_time(t)
                          * np.abs(u.iloc[:-1]))

        if self.b is not None:
            if current_and_past_volumes is None:
                raise SyntaxError(
                    "If you don't provide volumes you should set b to None")
            # we add 1 to the volumes to prevent 0 volumes error (trades are cancelled on 0 volumes)
            result += (np.abs(u.iloc[:-1])**exponent) @ (
                self.b._recursive_values_in_time(t) *
                sigma / ((current_and_past_volumes.iloc[-1] + 1) ** (exponent - 1)))

        assert not np.isnan(result)
        assert not np.isinf(result)

        return result

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):

        expression = 0
        if self.a is not None or self.pershare_cost is not None:
            expression += cp.abs(z[:-1]).T @ self.first_term_multiplier
            assert expression.is_convex()
        if self.b is not None:
            expression += (cp.abs(z[:-1]) ** (
                2 if self.exponent is None else self.exponent)
            ).T @ self.second_term_multiplier
            assert expression.is_convex()
        return expression


class StocksTransactionCost(TransactionCost):
    """A model for transaction costs of stocks.

    See pages 10-11 in `the book <https://stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_.
    We don't include the short-term alpha term `c` here because it
    can be expressed with a separate `ReturnsForecast` object. 

    :param a: linear cost, which multiplies the absolute value of each trade. This can model
        (half) the bid-ask spread, or any fee linear in the size of a trade.
    :type a: float or pd.Series or pd.DataFrame
    :param pershare_cost: per-share trade cost, amount of dollars paid for each share traded. 
    :type pershare_cost: float or pd.Series or pd.DataFrame
    :param b: coefficient of the non-linear term of the transaction cost model, which multiplies
        the estimated volatility for each stock (see the book).  
    :type b: float or pd.Series or pd.DataFrame
    :param window_sigma_est: we use an historical rolling standard deviation to estimate 
        the average size of the return on a stock on each day, and this multiplies the 
        second term of the transaction cost model.  See the paper for an explanation of the model. 
        Here you specify the length of the rolling window to use. If None (the default) it uses
        a length of 1 year (approximated with the data provided).
    :type window_sigma_est: int or None
    :param window_volume_est: length of the window for the mean of past volumes used as estimate
        of each period's volume. Has no effect on the simulator version of this which uses
        the actual volume. If None (the default) it uses a length of 1 year (approximated 
        with the data provided).
    :type window_volume_est: int
    :param exponent: exponent of the non-linear term, defaults (if set to ``None``) to 1.5 for
        the simulator version, and 2 for the optimization version (because it is more efficient
        numerically and the difference is small, you can change it if you want).
    :type exponent: float or None
    """

    def __init__(self, a=0., pershare_cost=0.005, b=1.0, window_sigma_est=None,
                 window_volume_est=None, exponent=1.5):

        super().__init__(a=a, pershare_cost=pershare_cost, b=b, window_sigma_est=window_sigma_est,
                         window_volume_est=window_volume_est, exponent=exponent)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the data and parameter estimator objects."""

import unittest
from pathlib import Path


import cvxpy as cp
import numpy as np
import pandas as pd


from cvxportfolio.forecast import HistoricalMeanReturn, HistoricalMeanError, HistoricalVariance, HistoricalFactorizedCovariance


class TestEstimators(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Load the data and initialize cvxpy vars."""
        # cls.sigma = pd.read_csv(Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
        cls.returns = pd.read_csv(
            Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
        # cls.volumes = pd.read_csv(Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
        cls.w_plus = cp.Variable(cls.returns.shape[1])
        cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
        cls.z = cp.Variable(cls.returns.shape[1])
        cls.N = cls.returns.shape[1]

    # def boilerplate(self, model):
    #     model._recursive_pre_evaluation(universe=self.returns.columns, backtest_times=self.returns.index)
    #     return model._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)

    def test_mean_update(self):
        forecaster = HistoricalMeanReturn()  # lastforcash=True)

        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            mean = forecaster._recursive_values_in_time(
                t=t, past_returns=past_returns)
            # print(mean)
            # self.assertTrue(mean[-1] == past_returns.iloc[-1,-1])
            self.assertTrue(np.allclose(
                mean, past_returns.iloc[:, :-1].mean()))

    def test_variance_update(self):
        forecaster = HistoricalVariance(kelly=False)

        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            var = forecaster._recursive_values_in_time(
                t=t, past_returns=past_returns)
            print(var)
            # self.assertTrue(mean[-1] == past_returns.iloc[-1,-1])
            self.assertTrue(np.allclose(var, past_returns.var(ddof=0)[:-1]))

    def test_meanerror_update(self):
        forecaster = HistoricalMeanError()

        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            val = forecaster._recursive_values_in_time(
                t=t, past_returns=past_returns)
            print(val)
            # self.assertTrue(mean[-1] == past_returns.iloc[-1,-1])
            self.assertTrue(np.allclose(val, past_returns.std(ddof=0)[
                            :-1] / np.sqrt(past_returns.count()[:-1])))

    def test_counts_matrix(self):
        forecaster = HistoricalFactorizedCovariance()  # kelly=True)
        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan
        returns.iloc[10:15, 10:20] = np.nan

        count_matrix = forecaster._get_count_matrix(returns)

        for indexes in [(1, 2), (4, 5), (1, 5), (7, 18), (7, 24), (1, 15), (13, 22)]:
            print(count_matrix.iloc[indexes[0], indexes[1]])
            print(len((returns.iloc[:, indexes[0]] *
                  returns.iloc[:, indexes[1]]).dropna()))
            self.assertTrue(np.isclose(count_matrix.iloc[indexes[0], indexes[1]],
                                       len((returns.iloc[:, indexes[0]] * returns.iloc[:, indexes[1]]).dropna())))

    def test_sum_matrix(self):
        forecaster = HistoricalFactorizedCovariance()  # kelly=True)
        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan
        returns.iloc[10:15, 10:20] = np.nan

        forecaster._recursive_values_in_time(
            t=pd.Timestamp('2022-01-01'), past_returns=returns)

        sum_matrix = forecaster.last_sum_matrix

        for indexes in [(1, 2), (4, 5), (1, 5), (7, 18), (7, 24), (1, 15), (13, 22)]:
            print()
            print(sum_matrix[indexes[0], indexes[1]])
            print((returns.iloc[:, indexes[0]] *
                  returns.iloc[:, indexes[1]]).sum())
            self.assertTrue(np.isclose(
                sum_matrix[indexes[0], indexes[1]],
                (returns.iloc[:, indexes[0]] *
                 returns.iloc[:, indexes[1]]).sum()
            ))

    def test_covariance_update(self):
        """Test covariance forecast estimator."""

        forecaster = HistoricalFactorizedCovariance()

        returns = pd.DataFrame(self.returns.iloc[:, :4], copy=True)
        returns.iloc[:20, 1] = np.nan
        returns.iloc[10:30, 0] = np.nan
        returns.iloc[25:40, 2] = np.nan

        def compute_Sigma(rets):
            res = np.zeros((3, 3))
            res[0, 0] = np.nanmean(rets.iloc[:, 0] * rets.iloc[:, 0])
            res[1, 1] = np.nanmean(rets.iloc[:, 1] * rets.iloc[:, 1])
            res[2, 2] = np.nanmean(rets.iloc[:, 2] * rets.iloc[:, 2])
            res[0, 1] = np.nanmean(rets.iloc[:, 0] * rets.iloc[:, 1])
            res[0, 2] = np.nanmean(rets.iloc[:, 0] * rets.iloc[:, 2])
            res[1, 2] = np.nanmean(rets.iloc[:, 1] * rets.iloc[:, 2])
            res[1, 0] = res[0, 1]
            res[2, 0] = res[0, 2]
            res[2, 1] = res[1, 2]
            return res

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            val = forecaster._recursive_values_in_time(
                t=t, past_returns=past_returns)
            Sigma = val @ val.T
            self.assertTrue(np.allclose(Sigma, compute_Sigma(past_returns)))

    def test_covariance_update_nokelly(self):
        """Test covariance forecast estimator.

        NOTE: due to a bug in pandas we can't test against pandas.DataFrame.cov, 
        see https://github.com/pandas-dev/pandas/issues/45814 . In fact with the
        current bug in pandas 
        ``past_returns.iloc[:,:-1].cov(ddof=0)`` returns ``past_returns.iloc[:,:-1].cov(ddof=1)``
        whenever there are missing values.
        """

        forecaster = HistoricalFactorizedCovariance(kelly=False)

        returns = pd.DataFrame(self.returns.iloc[:, :4], copy=True)
        returns.iloc[:20, 1] = np.nan
        returns.iloc[10:30, 0] = np.nan
        returns.iloc[25:40, 2] = np.nan

        def cov_ij(i, j, rets):
            i_nanmasker = np.zeros(len(rets))
            i_nanmasker[rets.iloc[:, i].isnull()] = np.nan
            i_nanmasker[~(rets.iloc[:, i].isnull())] = 1.
            j_nanmasker = np.zeros(len(rets))
            j_nanmasker[rets.iloc[:, j].isnull()] = np.nan
            j_nanmasker[~(rets.iloc[:, j].isnull())] = 1.
            print(i_nanmasker, j_nanmasker)
            return np.nanmean(rets.iloc[:, i] * rets.iloc[:, j]) - np.nanmean(rets.iloc[:, i] * j_nanmasker) * np.nanmean(rets.iloc[:, j] * i_nanmasker)

        def compute_Sigma(rets):
            res = np.zeros((3, 3))
            res[0, 0] = cov_ij(0, 0, rets)
            res[1, 1] = cov_ij(1, 1, rets)
            res[2, 2] = cov_ij(2, 2, rets)
            res[0, 1] = cov_ij(0, 1, rets)
            res[0, 2] = cov_ij(0, 2, rets)
            res[1, 2] = cov_ij(1, 2, rets)
            res[1, 0] = res[0, 1]
            res[2, 0] = res[0, 2]
            res[2, 1] = res[1, 2]
            return res

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            val = forecaster._recursive_values_in_time(
                t=t, past_returns=past_returns)
            Sigma = val @ val.T

            self.assertTrue(np.allclose(Sigma, compute_Sigma(past_returns)))
            # pandasSigma = past_returns.iloc[:,:-1].cov(ddof=0)
            # self.assertTrue(np.allclose(Sigma, pandasSigma))
            self.assertTrue(np.allclose(
                np.diag(Sigma), past_returns.iloc[:, :-1].var(ddof=0)))


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cvxportfolio __init__ module.

This module only republishes the api of a selection of cvxportfolio modules.
The __all__ attribute of each is used.
"""

__version__ = "0.4.8"

from .simulator import *
from .policies import *
from .constraints import *
from .costs import *
from .returns import *
from .risks import *
from .benchmark import *
from .result import *
from .hyperparameters import *



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DEPRECATED.

This module will be removed or heavily reformatted, most of it is unused.
The only parts that will remain are the FRED and YFinance interfaces, simplified,
and not meant to be accessed directly by users.
"""

from pathlib import Path
import yfinance as yf
import pandas as pd
import numpy as np
import sqlite3

from .estimator import DataEstimator

__all__ = ["YfinanceTimeSeries", "FredTimeSeries",
           # "FredRateTimeSeries", 
           "BASE_LOCATION"]

BASE_LOCATION = Path.home() / "cvxportfolio_data"


class BaseData:
    """Base class for Cvxportfolio database interface.

    Provides a back-end independent way to load and store
    pandas Series and DataFrames where the first index is a
    pandas Timestamp (or numpy datetime64). It also provides
    a systematic way to access external data sources via
    the network. We specialize it to storing data locally
    and downloading public time series of financial data.
    By emulating some of these classes you can interface
    cvxportfolio with other databases or other data sources.

    This interface is also used by cvxportfolio to store the
    data it generates, such as Backtest classes data, or
    estimators such as factor risk models.

    Cvxportfolio uses in-memory data whenever possible, and
    in particular never uses BaseData methods during a backtest.
    This ensures thread safety and allows us to use simple
    local databases such as sqlite (or even flat csv files!).
    Cvxportfolio loads data from the database before (possibly parallel)
    backtesting, and stores it after backtesting. So, only one process
    at a time accesses this class' methods. If you write custom callbacks
    that are invoked during backtests, such as callables inside
    DataEstimator, you should most probably not use cvxportfolio.data methods
    inside them.

    LIMITATIONS:
        - columns names should be strings in order to work with all current
          data storage backends. If you create a DataFrame from a numpy array
          without specifying column names they will default to integers (0, 1, ...).
          If you store and load it back you will may get string column names ('0', '1', ...),
          depending on the backend.
        - the first level of the index should be a pandas Timestamp or equivalently
          numpy datetime64
        - you can only store sql-friendly data: integers, floats (including `np.nan`),
          datetime (including `np.datetime64('NaT')`), and simple alphanumeric strings
          (i.e., without commas or quote marks).
          If you need to store more complex python objects, such as the string
            "{'parameter1':3.0, 'parameter2': pd.Timestamp('2022-01-01')}",
          you will have to check that it works with the backend you use
         (it probably would not not with csv).
    """

    def load_raw(self, symbol):
        """Load raw data from database."""
        raise NotImplementedError

    def load(self, symbol):
        """Load data from database using `self.preload` function to process it."""
        return self.preload(self.load_raw(symbol))

    def store(self, symbol, data):
        """Store data in database."""
        raise NotImplementedError

    def download(self, symbol, current):
        """Download data from external source."""
        raise NotImplementedError

    def update(self, symbol):
        """Update current stored data for symbol."""
        current = self.load_raw(symbol)
        updated = self.download(symbol, current)
        self.store(symbol, updated)

    def update_and_load(self, symbol):
        """Update current stored data for symbol and load it.

        DEPRECATED: update and load functionalities have been separated.
        """
        current = self.load_raw(symbol)
        updated = self.download(symbol, current)
        self.store(symbol, updated)
        # return self.preload(updated)
        return self.load(symbol)

    def preload(self, data):
        """Prepare data to serve to the user."""
        return data


class YfinanceBase(BaseData):
    """Base class for the Yahoo Finance interface.

    This should not be used directly unless you know what you're doing.
    """

    @staticmethod
    def internal_process(data):
        """Manipulate yfinance data for better storing."""

        # nan-out nonpositive prices
        data.loc[data["Open"] <= 0, 'Open'] = np.nan
        data.loc[data["Close"] <= 0, "Close"] = np.nan
        data.loc[data["High"] <= 0, "High"] = np.nan
        data.loc[data["Low"] <= 0, "Low"] = np.nan
        data.loc[data["Adj Close"] <= 0, "Adj Close"] = np.nan

        # nan-out negative volumes
        data.loc[data["Volume"] < 0, 'Volume'] = np.nan

        intraday_logreturn = np.log(data["Close"]) - np.log(data["Open"])
        close_to_close_logreturn = np.log(data["Adj Close"]).diff().shift(-1)
        open_to_open_logreturn = (
            close_to_close_logreturn + intraday_logreturn -
            intraday_logreturn.shift(-1)
        )
        data["Return"] = np.exp(open_to_open_logreturn) - 1
        del data["Adj Close"]
        # eliminate last period's intraday data
        data.loc[data.index[-1], ["High", "Low",
                                  "Close", "Return", "Volume"]] = np.nan
        return data

    def download(self, symbol, current=None, overlap=5, grace_period='5d', **kwargs):
        """Download single stock from Yahoo Finance.

        If data was already downloaded we only download
        the most recent missing portion.

        Args:

            symbol (str): yahoo name of the instrument
            current (pandas.DataFrame or None): current data present locally
            overlap (int): how many lines of current data will be overwritten
                by newly downloaded data
            kwargs (dict): extra arguments passed to yfinance.download

        Returns:
            updated (pandas.DataFrame): updated DataFrame for the symbol
        """
        if (current is None) or (len(current) < overlap):
            updated = yf.download(symbol, progress=False, **kwargs)
            return self.internal_process(updated)
        else:
            if (pd.Timestamp.today() - current.index[-1]) < pd.Timedelta(grace_period):
                return current
            new = yf.download(symbol, progress=False,
                              start=current.index[-overlap], **kwargs)
            new = self.internal_process(new)
            return pd.concat([current.iloc[:-overlap], new])

    @staticmethod
    def preload(data):
        """Prepare data for use by Cvxportfolio.

        We drop the 'Volume' column expressed in number of stocks
        and replace it with 'ValueVolume' which is an estimate
        of the (e.g., US dollar) value of the volume exchanged
        on the day.
        """
        data["ValueVolume"] = data["Volume"] * data["Open"]
        del data["Volume"]
        # remove infty values
        data.iloc[:, :] = np.nan_to_num(
            data.values, copy=True, nan=np.nan, posinf=np.nan, neginf=np.nan)
        # remove extreme values
        # data.loc[data["Return"] < -.99, "Return"] = np.nan
        # data.loc[data["Return"] > .99, "Return"] = np.nan
        return data


class BaseDataStore(BaseData):
    """Base class for data storage systems.

    Attributes:
        base_location (str or None): location of storage.
    """

    base_location = None


# class SqliteDataStore(BaseDataStore):
#     """Local sqlite3 database using python standard library.
#
#     Args:
#         location (pathlib.Path or None): pathlib.Path base location of the databases
#             directory or, if None, use ":memory:" for storing in RAM instead. Default
#             is ~/cvxportfolio/
#     """
#
#     def __init__(self, base_location=BASE_LOCATION):
#         """Initialize sqlite connection and if necessary create database."""
#         self.base_location = base_location
#
#     def __open__(self):
#         """Open database connection."""
#         if self.base_location is None:
#             self.connection = sqlite3.connect(":memory:")
#         else:
#             self.connection = sqlite3.connect(
#                 (self.base_location / self.__class__.__name__).with_suffix(".sqlite"))
#
#     def __close__(self):
#         """Close database connection."""
#         self.connection.close()
#
#     def store(self, symbol, data):
#         """Store Pandas object to sqlite.
#
#         We separately store dtypes for data consistency and safety.
#
#         Limitations: if your pandas object's index has a name it will be lost,
#             the index is renamed 'index'.
#         """
#         self.__open__()
#         exists = pd.read_sql_query(
#             f"SELECT name FROM sqlite_master WHERE type='table' AND name='{symbol}'",
#             self.connection,
#         )
#         if len(exists):
#             res = self.connection.cursor().execute(f"DROP TABLE '{symbol}'")
#             res = self.connection.cursor().execute(
#                 f"DROP TABLE '{symbol}___dtypes'")
#             self.connection.commit()
#
#         if hasattr(data.index, "levels"):
#             data.index = data.index.set_names(
#                 ["index"] +
#                 [f"___level{i}" for i in range(1, len(data.index.levels))]
#             )
#             data = data.reset_index().set_index("index")
#         else:
#             data.index.name = "index"
#
#         data.to_sql(f"{symbol}", self.connection)
#         pd.DataFrame(data).dtypes.astype("string").to_sql(
#             f"{symbol}___dtypes", self.connection
#         )
#         self.__close__()
#
#     def load_raw(self, symbol):
#         """Load Pandas object with datetime index from sqlite.
#
#         If data is not present in in the database, return None.
#         """
#         try:
#             self.__open__()
#             dtypes = pd.read_sql_query(
#                 f"SELECT * FROM {symbol}___dtypes",
#                 self.connection,
#                 index_col="index",
#                 dtype={"index": "str", "0": "str"},
#             )
#             tmp = pd.read_sql_query(
#                 f"SELECT * FROM {symbol}",
#                 self.connection,
#                 index_col="index",
#                 parse_dates="index",
#                 dtype=dict(dtypes["0"]),
#             )
#             self.__close__()
#             multiindex = []
#             for col in tmp.columns:
#                 if col[:8] == "___level":
#                     multiindex.append(col)
#                 else:
#                     break
#             if len(multiindex):
#                 multiindex = [tmp.index.name] + multiindex
#                 tmp = tmp.reset_index().set_index(multiindex)
#             return tmp.iloc[:, 0] if tmp.shape[1] == 1 else tmp
#         except pd.errors.DatabaseError:
#             return None


class PickleStore(BaseDataStore):
    """Pickle data store for pandas Series and DataFrames.

    Args:
        base_location (pathlib.Path): filesystem directory where to store files.

    """

    # base_location = BASE_LOCATION

    @property
    def location(self):
        return self.base_location / self.__class__.__name__

    def __init__(self, base_location=BASE_LOCATION):
        self.base_location = base_location

    def __create_if_not_existent(self):
        if not self.location.is_dir():
            self.location.mkdir(parents=True)
            print(f"Created folder at {self.location}")

    def load_raw(self, symbol, **kwargs):
        """Load raw data from local store."""
        try:
            return pd.read_pickle(self.location / f"{symbol}.pickle")
        except FileNotFoundError:
            return None

    def store(self, symbol, data, **kwargs):
        """Store data locally."""
        self.__create_if_not_existent()
        data.to_pickle(self.location / f"{symbol}.pickle")


class LocalDataStore(BaseDataStore):
    """Local data store for pandas Series and DataFrames.

    Args:
        base_location (pathlib.Path): filesystem directory where to store files.

    """

    # base_location = Path.home() / "cvxportfolio"

    @property
    def location(self):
        return self.base_location / self.__class__.__name__

    def __init__(self, base_location=BASE_LOCATION):
        self.base_location = base_location

    def __create_if_not_existent(self):
        if not self.location.is_dir():
            self.location.mkdir(parents=True)
            print(f"Created folder at {self.location}")

    def load_raw(self, symbol, **kwargs):
        """Load raw data from local store."""
        try:
            try:
                multiindex_types = pd.read_csv(
                    self.location /
                    f"{symbol}___multiindex_dtypes.csv",
                    index_col=0)["0"]
            except FileNotFoundError:
                multiindex_types = ["datetime64[ns]"]
            dtypes = pd.read_csv(
                self.location / f"{symbol}___dtypes.csv",
                index_col=0,
                dtype={"index": "str", "0": "str"},
            )
            dtypes = dict(dtypes["0"])
            new_dtypes = {}
            parse_dates = []
            for i, level in enumerate(multiindex_types):
                if level == "datetime64[ns]":
                    parse_dates.append(i)
            for i, el in enumerate(dtypes):
                if dtypes[el] == "datetime64[ns]":
                    parse_dates += [i + len(multiindex_types)]
                else:
                    new_dtypes[el] = dtypes[el]

            # raise Exception
            tmp = pd.read_csv(
                self.location / f"{symbol}.csv",
                index_col=list(range(len(multiindex_types))),
                parse_dates=parse_dates,
                **kwargs,
                dtype=new_dtypes,
            )
            return tmp.iloc[:, 0] if tmp.shape[1] == 1 else tmp
        except FileNotFoundError:
            return None

    def store(self, symbol, data, **kwargs):
        """Store data locally."""
        self.__create_if_not_existent()
        if hasattr(data.index, "levels"):
            pd.DataFrame(data.index.dtypes).astype("string").to_csv(
                self.location / f"{symbol}___multiindex_dtypes.csv"
            )
        pd.DataFrame(data).dtypes.astype("string").to_csv(
            self.location / f"{symbol}___dtypes.csv"
        )
        data.to_csv(self.location / f"{symbol}.csv", **kwargs)


class FredBase(BaseData):
    """Base class for FRED data access."""

    URL = "https://fred.stlouisfed.org/graph/fredgraph.csv"

    def _download(self, symbol):
        return pd.read_csv(self.URL + f'?id={symbol}', index_col=0, parse_dates=[0])[symbol]

    def download(self, symbol="DFF", current=None, grace_period='5d'):
        """Download or update pandas Series from FRED. 

        If already downloaded don't change data stored locally and only
        add new entries at the end. 

        Additionally, we allow for a `grace period`, if the data already downloaded
        has a last entry not older than the grace period, we don't download new data.
        """
        if current is None:
            return self._download(symbol)
        else:
            if (pd.Timestamp.today() - current.index[-1]) < pd.Timedelta(grace_period):
                return current

            new = self._download(symbol)
            new = new.loc[new.index > current.index[-1]]

            if new.empty:
                return current

            assert new.index[0] > current.index[-1]
            return pd.concat([current, new])


class RateBase(BaseData):
    """Manipulate rate data from percent annualized to daily."""

    trading_days = 252

    def preload(self, data):
        return np.exp(np.log(1 + data / 100) / self.trading_days) - 1


class Yfinance(YfinanceBase, LocalDataStore):

    """Yahoo Finance data interface using local data store.

    Args:
        base_location (pathlib.Path): filesystem directory where to store files.
    """

    def update_and_load(self, symbol):
        """Update data for symbol and load it."""
        return super().update_and_load(symbol)


class FredRate(FredBase, RateBase, PickleStore):
    """Load and store FRED rates like DFF."""

    pass


class YfinanceTimeSeries(DataEstimator, YfinanceBase, PickleStore):

    def __init__(self, symbol, use_last_available_time=False, base_location=BASE_LOCATION):
        self.symbol = symbol
        self.base_location = base_location
        self.use_last_available_time = use_last_available_time

    def _recursive_pre_evaluation(self, *args, **kwargs):
        self.data = self.update_and_load(self.symbol)


class FredTimeSeries(DataEstimator, FredBase, PickleStore):

    def __init__(self, symbol, use_last_available_time=False, base_location=BASE_LOCATION):
        self.symbol = symbol
        self.base_location = base_location
        self.use_last_available_time = use_last_available_time

    def _recursive_pre_evaluation(self, *args, **kwargs):
        self.data = self.update_and_load(self.symbol)


# class FredRateTimeSeries(DataEstimator, FredBase, RateBase, PickleStore):
#
#     def __init__(self, symbol, use_last_available_time=False, base_location=BASE_LOCATION):
#         self.symbol = symbol
#         self.base_location = base_location
#         self.use_last_available_time = use_last_available_time
#
#     def _recursive_pre_evaluation(self, *args, **kwargs):
#         self.data = self.update_and_load(self.symbol)


class TimeSeries(DataEstimator):
    """Class for time series data managed by Cvxportfolio.

    Args:
        symbol (str): name of the time series, such as 'AAPL',
            '^VIX', or 'DFF'.
        source (str or BaseData): data source to use. Currently we
            support 'yahoo', equivalent to `cvxportfolio.YfinanceBase`,
            and 'fred', equivalent to `cvxportfolio.FredBase`. If you
            implement your own you should define the `download` and
            optionally `preload` methods. Default is 'yahoo'.
        storage (str or BaseData): storage backend to use. Currently we
            support 'sqlite', equivalent to `cvxportfolio.SqliteDataStore`,
            and 'csv', equivalent to `cvxportfolio.LocalDataStore`. If you
            implement your own you should define the `store` and
            `load_raw` methods. Default is 'sqlite'.
        use_last_available_time (bool): as in `cvxportfolio.DataEstimator`
        base_location (pathlib.Path or None): base location for the data storage.


    """

    def __init__(
        self,
        symbol,
        source="yahoo",
        storage="pickle",
        use_last_available_time=False,
        base_location=None,
    ):
        self.symbol = symbol
        if isinstance(source, str) and source == "yahoo":
            source = YfinanceBase
        if isinstance(source, str) and source == "fred":
            source = FredRate

        # from
        # https://stackoverflow.com/questions/11042424/adding-base-class-to-existing-object-in-python
        cls = self.__class__
        self.__class__ = cls.__class__(
            cls.__name__ + "With" + source.__name__, (cls, source), {}
        )

        # if isinstance(storage, str) and storage == "sqlite":
        #     storage = SqliteDataStore
        if isinstance(storage, str) and storage == "csv":
            storage = LocalDataStore
        if isinstance(storage, str) and storage == "pickle":
            storage = PickleStore

        cls = self.__class__
        self.__class__ = cls.__class__(
            cls.__name__ + "With" + storage.__name__, (cls, storage), {}
        )

        self.base_location = base_location
        self.use_last_available_time = use_last_available_time
        self.universe_maybe_noncash = None # fix, but we should retire this class

    def _recursive_pre_evaluation(self, *args, **kwargs):
        self.data = self.update_and_load(self.symbol)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for some utils.py functions."""

import unittest

import cvxpy as cp
import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.utils import *


class TestUtils(unittest.TestCase):

    def test_hasher(self):
        a = np.array([1, 2, 3])
        re = repr_numpy_pandas(a)
        print(re)
        b = np.array([[1, 2, 3]])
        re1 = repr_numpy_pandas(b)
        print(re1)
        self.assertTrue(re == re1)

        a = pd.Series([1, 2, 3], ['2020-01-01', '2021-01-01', '2022-01-01'])
        re = repr_numpy_pandas(a)
        print(re)
        b = pd.Series(a.values, a.index)
        re1 = repr_numpy_pandas(b)
        print(re1)
        self.assertTrue(re == re1)

        # multiindexed df
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        data.columns = ["one", "two", "tre", "quattro",
                        "cinque", "sei", "sette", "otto", "nove", "dieci"]

        re = repr_numpy_pandas(data)
        print(re)

        data1 = pd.DataFrame(data.values,
                             index=data.index, columns=data.columns)

        re1 = repr_numpy_pandas(data1)
        print(re1)
        self.assertTrue(re == re1)


if __name__ == '__main__':

    unittest.main()



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the data interfaces."""

import unittest
import shutil
import tempfile
from pathlib import Path

import numpy as np
import pandas as pd

from cvxportfolio.data import (
    YfinanceBase,
    LocalDataStore,
    Yfinance,
    FredBase,
    FredRate,
    # SqliteDataStore,
    TimeSeries,
    PickleStore,
)


class TestData(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Initialize data directory."""
        cls.datadir = Path(tempfile.mkdtemp())
        print('created', cls.datadir)

    @classmethod
    def tearDownClass(cls):
        """Remove data directory."""
        print('removing', cls.datadir)
        shutil.rmtree(cls.datadir)

    def test_time_series(self):
        ts = TimeSeries("ZM", base_location=self.datadir)
        assert not hasattr(ts, "data")
        ts._recursive_pre_evaluation()
        assert np.all(
            ts._recursive_values_in_time(
                pd.Timestamp("2023-04-11"), "foo", bar=None)
            == ts.data.loc["2023-04-11"]
        )

    def test_yfinance_download(self):
        """Test YfinanceBase."""

        data = YfinanceBase().download("AAPL", start="2023-04-01", end="2023-04-15")
        print(data)
        print(data.loc["2023-04-10"]["Return"])
        print(data.loc["2023-04-11", "Open"] /
              data.loc["2023-04-10", "Open"] - 1)
        self.assertTrue(np.isclose(
            data.loc["2023-04-10", "Return"],
            data.loc["2023-04-11", "Open"] /
            data.loc["2023-04-10", "Open"] - 1,
        ))
        self.assertTrue(np.isnan(data.iloc[-1]["Close"]))

    def test_fred(self):
        store = FredRate(self.datadir)
        data = store.update_and_load("DFF")
        print(data)
        self.assertTrue(np.isclose((1 + data["2023-04-10"]) **
                                   store.trading_days, 1 + 4.83 / 100))

    def test_fred_base(self):
        data = FredBase().download("DFF")
        self.assertTrue(data["2023-04-06"] == 4.83)
        self.assertTrue(data.index[0] == pd.Timestamp("1954-07-01"))
        print(data)
        olddata = data.iloc[:-123]
        newdata = FredBase().download("DFF", olddata)
        self.assertTrue(np.all(data == newdata))

    def test_yfinance(self):
        """Test yfinance ability to store and retrieve."""

        store = Yfinance(self.datadir)
        data = store.update_and_load("ZM")

        print(data)

        self.assertTrue(np.isclose(
            data.loc["2023-04-05", "Return"],
            data.loc["2023-04-06", "Open"] /
            data.loc["2023-04-05", "Open"] - 1,
        ))

        data1 = store.update_and_load("ZM")
        print(data1)

        self.assertTrue(np.isnan(data1.iloc[-1]["Close"]))

        print((data1.iloc[: len(data) - 1].Return -
              data.iloc[:-1].Return).describe().T)

        self.assertTrue(np.allclose(
            data1.loc[data.index[:-1]].Return, data.iloc[:-1].Return))

    # def test_sqlite3_store_series(self):
    #     """Test storing and retrieving of a Series with datetime index."""
    #     self.base_test_series(SqliteDataStore, self.datadir)

    def test_local_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self.base_test_series(LocalDataStore, self.datadir)

    def test_pickle_store_series(self):
        """Test storing and retrieving of a Series with datetime index."""
        self.base_test_series(PickleStore, self.datadir)

    # def test_sqlite3_store_dataframe(self):
    #     """Test storing and retrieving of a DataFrame with datetime index."""
    #     self.base_test_dataframe(SqliteDataStore, self.datadir)

    def test_local_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_dataframe(LocalDataStore, self.datadir)

    def test_pickle_store_dataframe(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_dataframe(PickleStore, self.datadir)

    def test_local_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_multiindex(LocalDataStore, self.datadir)

    # def test_sqlite3_store_multiindex(self):
    #     """Test storing and retrieving of a DataFrame with datetime index."""
    #     self.base_test_multiindex(SqliteDataStore, self.datadir)

    def test_pickle_store_multiindex(self):
        """Test storing and retrieving of a DataFrame with datetime index."""
        self.base_test_multiindex(PickleStore, self.datadir)

    def base_test_series(self, storeclass, *args, **kwargs):
        """Test storing and retrieving of a Series with datetime index."""
        store = storeclass(*args, **kwargs)
        for data in [
            pd.Series(
                0.0,
                pd.date_range(
                    "2020-01-01",
                    "2020-01-10"),
                name="prova1"),
            pd.Series(3, pd.date_range("2020-01-01",
                      "2020-01-10"), name="prova2"),
            pd.Series(
                "ciao", pd.date_range("2020-01-01", "2020-01-02", freq="H"), name="prova3"
            ),
            # test datetime conversion
            pd.Series(
                pd.date_range("2022-01-01", "2022-01-02", freq="H"),
                pd.date_range("2020-01-01", "2020-01-02", freq="H"),
                name="prova4",
            ),
            # test overwrite
            pd.Series(
                pd.date_range("2023-01-01", "2023-01-02", freq="H"),
                pd.date_range("2020-01-01", "2020-01-02", freq="H"),
                name="prova4",
            ),
        ]:
            print(data)
            print(data.index.dtype)
            print(data.dtypes)

            store.store(data.name, data)

            data1 = store.load(data.name)
            print(data1)
            print(data1.index.dtype)
            print(data1.dtypes)

            self.assertTrue(data.name == data1.name)
            self.assertTrue(all(data == data1))
            self.assertTrue(all(data.index == data1.index))
            self.assertTrue(data.dtypes == data1.dtypes)

        self.assertTrue(store.load("blahblah") is None)

    def base_test_dataframe(self, storeclass, *args, **kwargs):
        """Test storing and retrieving of a DataFrame with datetime index."""
        store = storeclass(*args, **kwargs)
        index = pd.date_range("2020-01-01", "2020-01-02", freq="H")
        data = {
            "one": range(len(index)),
            "two": np.arange(len(index)) / 19.0,
            "three": ["hello"] * len(index),
            "four": [np.nan] * len(index),
        }

        data["two"][2] = np.nan
        data = pd.DataFrame(data, index=index)
        print(data)
        print(data.index.dtype)
        print(data.dtypes)

        store.store("example", data)
        data1 = store.load("example")
        print(data1)
        print(data1.index.dtype)
        print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.dtypes == data1.dtypes))

    def base_test_multiindex(self, storeclass, *args, **kwargs):
        """Test storing and retrieving of a Series or DataFrame with multi-index."""
        store = storeclass(*args, **kwargs)

        # second level is object
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        data.columns = ["one", "two", "tre", "quattro",
                        "cinque", "sei", "sette", "otto", "nove", "dieci"]

        print(data.index)
        print(data)
        print(data.index.dtype)
        print(data.dtypes)

        store.store("example", data)
        data1 = store.load("example")

        print(data1.index)
        print(data1)
        print(data1.index.dtype)
        print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.index.dtypes == data1.index.dtypes))
        self.assertTrue(all(data.dtypes == data1.dtypes))

        # second level is timestamp
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = pd.date_range("2022-01-01", "2022-01-03")
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        data.columns = [
            "one",
            "two",
            "tre",
            "quattro",
            "cinque",
            "sei",
            "sette",
            "otto",
            "nove",
            "dieci",
        ]

        print(data.index)
        print(data)
        print(data.index.dtype)
        print(data.dtypes)

        store.store("example", data)
        data1 = store.load("example")

        print(data1.index)
        print(data1)
        print(data1.index.dtype)
        print(data1.dtypes)

        self.assertTrue(all(data == data1))
        self.assertTrue(all(data.index == data1.index))
        self.assertTrue(all(data.index.dtypes == data1.index.dtypes))
        self.assertTrue(all(data.dtypes == data1.dtypes))


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib

import pandas as pd
import numpy as np

TRUNCATE_REPR_HASH = 10  # probability of conflict is 1e-16


__all__ = ['periods_per_year', 'resample_returns',
           'flatten_heterogeneous_list', 'repr_numpy_pandas']


def periods_per_year(idx):
    """Given a datetime pandas index return the periods per year."""
    return int(np.round(len(idx) / ((idx[-1] - idx[0]) /
                                    pd.Timedelta('365.24d'))))


def resample_returns(returns, periods):
    """Resample returns expressed over number of periods to single period."""
    return np.exp(np.log(1 + returns) / periods) - 1


def flatten_heterogeneous_list(li):
    """[1, 2, 3, [4, 5]] -> [1, 2, 3, 4, 5]"""
    return sum(([el] if not hasattr(el, '__iter__')
                else el for el in li), [])


def hash_(array_like):
    """Hash np.array."""
    return hashlib.sha256(
        bytes(str(list(array_like.flatten())), 'utf-8')).hexdigest()[
            :TRUNCATE_REPR_HASH]


def repr_numpy_pandas(nppd):
    """Unique repr of a numpy or pandas object."""
    if isinstance(nppd, np.ndarray):
        return f'np.array(hash={hash_(nppd)})'
    # if isinstance(nppd, np.matrix):
    #     return f'np.array(hash={hash_(nppd.A)})'
    if isinstance(nppd, pd.Series):
        return f'pd.Series(hash_values={hash_(nppd.values)}, ' \
            + f'hash_index={hash_(nppd.index.to_numpy())})'
    if isinstance(nppd, pd.DataFrame):
        return f'pd.DataFrame(hash_values={hash_(nppd.values)}, ' \
            + f'hash_index={hash_(nppd.index.to_numpy())}, ' \
            + f'hash_columns={hash_(nppd.columns.to_numpy())})'
    raise NotImplementedError('The provided data type is not supported.')



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import collections
import numpy as np
import pandas as pd
import copy

from .estimator import Estimator
from .utils import periods_per_year
import matplotlib.pyplot as plt

__all__ = ['BacktestResult']


# def getFiscalQuarter(dt):
#     """Convert a time to a fiscal quarter."""
#     year = dt.year
#     quarter = (dt.month - 1) // 3 + 1
#     return "Q%i %s" % (quarter, year)


class BacktestResult(Estimator):
    """Holds the data from a Backtest and producs metrics and plots."""

    def __init__(self, universe, backtest_times, costs):
        """Initialization of backtest result."""
        self.h = pd.DataFrame(index=backtest_times,
                              columns=universe, dtype=float)
        self.u = pd.DataFrame(index=backtest_times,
                              columns=universe, dtype=float)
        self.z = pd.DataFrame(index=backtest_times,
                              columns=universe, dtype=float)
        self.costs = {cost.__class__.__name__: pd.Series(
            index=backtest_times, dtype=float) for cost in costs}
        self.policy_times = pd.Series(index=backtest_times, dtype=float)
        self.simulator_times = pd.Series(index=backtest_times, dtype=float)

    @property
    def PPY(self):
        return periods_per_year(self.h.index)

    @property
    def v(self):
        """The value of the portfolio over time."""
        return self.h.sum(axis=1)

    @property
    def profit(self):
        """The profit made, in dollars."""
        return self.v.iloc[-1] - self.v.iloc[0]

    @property
    def w(self):
        """The weights of the portfolio in time."""
        return (self.h.T / self.v).T

    @property
    def leverage(self):
        """Portfolio leverage"""
        return np.abs(self.w.iloc[:, :-1]).sum(1)

    @property
    def volatility(self):
        """The annualized, realized portfolio volatility."""
        return np.sqrt(self.PPY) * np.std(self.returns)

    @property
    def mean_return(self):
        """The annualized mean portfolio return."""
        return self.PPY * np.mean(self.returns)

    @property
    def sharpe_ratio(self):
        return (np.sqrt(self.PPY)
                * np.mean(self.excess_returns)
                / np.std(self.excess_returns)
                )

    @property
    def returns(self):
        """The returns R_t = (v_{t+1}-v_t)/v_t"""
        val = self.v
        return pd.Series(
            data=val.values[1:] / val.values[:-1] - 1, index=val.index[:-1]
        )

    @property
    def excess_returns(self):
        return self.returns - self.cash_returns

    @property
    def growth_rates(self):
        """The growth rate log(v_{t+1}/v_t)"""
        return np.log(self.returns + 1)

    @property
    def excess_growth_rates(self):
        """The growth rate log(v_{t+1}/v_t)"""
        return np.log(self.excess_returns + 1)

    @staticmethod
    def _print_growth_rate(gr):
        """Transform growth rate into return and pretty-print it.

        Either prints in basis points, percentage, or
        multiplication.
        """
        ret = np.exp(gr)-1
        if np.abs(ret) < 0.005:
            return f'{int(ret*1E4):d}bp'
        if np.abs(gr) < 1:
            return f'{ret*100:.2f}%'
        return f'{1+ret:.1f}X'

    # @property
    # def annual_growth_rate(self):
    #     """The annualized growth rate PPY/T sum_{t=1}^T log(v_{t+1}/v_t)"""
    #     return self.growth_rates.mean() * self.PPY
    #
    # @property
    # def annual_return(self):
    #     """The annualized return in percent."""
    #     ret = self.growth_rates
    #     return self._growth_to_return(ret.mean())
    #
    # def _growth_to_return(self, growth_rates):
    #     """Convert growth to annualized percentage return."""
    #     return 100 * (np.exp(self.PPY * growth) - 1)

    # def get_quarterly_returns(self, benchmark=None):
    #     """The annualized returns for each fiscal quarter."""
    #     ret = self.growth_rates
    #     quarters = ret.groupby(getFiscalQuarter).aggregate(np.mean)
    #     return self._growth_to_return(quarters)
    #
    # def get_best_quarter(self, benchmark=None):
    #     ret = self.get_quarterly_returns(benchmark)
    #     return (ret.argmax(), ret.max())
    #
    # def get_worst_quarter(self, benchmark=None):
    #     ret = self.get_quarterly_returns(benchmark)
    #     return (ret.argmin(), ret.min())

    @property
    def turnover(self):
        """Turnover ||u_t||_1/(2*v_t)"""
        return np.abs(self.u.iloc[:, :-1]).sum(axis=1) / (2*self.v.loc[self.u.index])

    @property
    def trading_days(self):
        """The fraction of days with nonzero turnover."""
        return (self.turnover.values > 0).sum() / self.turnover.size

    @property
    def drawdown(self):
        return -(1 - (self.v / self.v.cummax()))

    def plot(self, show=True, how_many_weights=7):
        """Make plots."""

        # value
        self.v.plot(figsize=(12, 5), label='Multi Period Optimization')
        plt.ylabel('USD')
        plt.yscale('log')
        plt.title('Total value of the portfolio in time')

        # weights
        biggest_weights = np.abs(self.w).mean(
        ).sort_values().iloc[-how_many_weights:].index
        self.w[biggest_weights].plot()
        plt.title('Largest weights of the portfolio in time')

        if show:
            plt.show()

    def __repr__(self):
        data = collections.OrderedDict({
            "Number of periods": self.u.shape[0],
            "Initial timestamp": self.h.index[0],
            "Universe size": self.h.shape[1],
            "Final timestamp": self.h.index[-1],
            "Total profit (PnL)": self.profit,
            "Initial portfolio value": self.v.iloc[0],
            "Final portfolio value": self.v.iloc[-1],
            # returns
            "Annualized absolute return (%)": 100 * self.mean_return,
            "Annualized absolute risk (%)": 100 * self.volatility,
            "Annualized excess return (%)": self.excess_returns.mean() * 100 * self.PPY,
            "Annualized excess risk (%)": self.excess_returns.std() * 100 * np.sqrt(self.PPY),
            # growth rates
            "Per-period absolute growth rate": self._print_growth_rate(self.growth_rates.mean()),
            "Per-period excess growth rate": self._print_growth_rate(self.excess_growth_rates.mean()),
            # stats
            "Sharpe ratio": self.sharpe_ratio,
            "Worst drawdown (%)": self.drawdown.min() * 100,
            "Average drawdown (%)": self.drawdown.mean() * 100,
            "Per-period Turnover (%)": self.turnover.mean() * 100,
            "Annualized Turnover (%)": self.turnover.mean() * 100 * self.PPY,

            "Average leverage (%)": self.leverage.mean() * 100,
            "Max leverage (%)": self.leverage.max() * 100})

        data.update(collections.OrderedDict({f'Average of {cost} per period (bp)': (
            self.costs[cost]/self.v).mean()*1E4 for cost in self.costs}))

        data.update(collections.OrderedDict(
            {"Average policy time (sec)": self.policy_times.mean(),
             "Average simulator time (sec)": self.simulator_times.mean()
             }))

        return 'Backtest Result:\n' + pd.Series(data=data).to_string(float_format="{:,.3f}".format)



from setuptools import setup

with open('README.md') as f:
    long_descr = ''.join(f.readlines())

setup(
    name='cvxportfolio',
    version='0.4.8',
    author='Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.',
    maintainer='Enzo Busseti',
    author_email='enzo.busseti@gmail.com',
    packages=['cvxportfolio',
              'cvxportfolio.tests'],
    package_dir={'cvxportfolio': 'cvxportfolio'},
    package_data={'cvxportfolio': [
        'tests/returns.csv', 'tests/sigmas.csv', 'tests/volumes.csv']},
    url='https://cvxportfolio.readthedocs.io',
    license='Apache 2.0',
    description='Portfolio optimization.',
    install_requires=["pandas",
                      "numpy",
                      "scipy",
                      "matplotlib",
                      "yfinance",
                      "cvxpy",
                      "multiprocess"
                      ],
    long_description=long_descr,
    long_description_content_type='text/markdown',
)



# # Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# #    http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
#
# import unittest
# from pathlib import Path
#
# import numpy as np
# import pandas as pd
# import cvxpy as cp
# import cvxportfolio as cvx
#
#
# class BaseTestClass(unittest.TestCase):
#
#     @classmethod
#     def setUpClass(cls):
#         """Load the data and initialize cvxpy vars."""
#         cls.sigma = pd.read_csv(
#             Path(__file__).parent / "sigmas.csv", index_col=0, parse_dates=[0])
#         cls.returns = pd.read_csv(
#             Path(__file__).parent / "returns.csv", index_col=0, parse_dates=[0])
#         cls.volumes = pd.read_csv(
#             Path(__file__).parent / "volumes.csv", index_col=0, parse_dates=[0])
#         cls.w_plus = cp.Variable(cls.returns.shape[1])
#         cls.w_plus_minus_w_bm = cp.Variable(cls.returns.shape[1])
#         cls.z = cp.Variable(cls.returns.shape[1])
#         cls.N = cls.returns.shape[1]
#
#     def boilerplate(self, model):
#         model._recursive_pre_evaluation(
#             universe=self.returns.columns, backtest_times=self.returns.index)
#         return model._compile_to_cvxpy(self.w_plus, self.z, self.w_plus_minus_w_bm)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .estimator import DataEstimator
import logging
import warnings

import scipy.linalg

import cvxpy as cp
import numpy as np
import pandas as pd

from .costs import BaseCost
from .forecast import HistoricalVariance, HistoricalFactorizedCovariance

logger = logging.getLogger(__name__)


__all__ = [
    "FullCovariance",
    "DiagonalCovariance",
    "FactorModelCovariance",
    "RiskForecastError",
    "WorstCaseRisk",
]


class BaseRiskModel(BaseCost):
    pass


class FullCovariance(BaseRiskModel):
    """Quadratic risk model with full covariance matrix.

    :param Sigma: DataFrame of covariance matrices
        supplied by the user, or None if fitting from the past data.
        The DataFrame can either represents a single constant covariance matrix
        or one for each point in time.
    :type Sigma: pandas.DataFrame or None


    """
    # r"""Quadratic risk model with full covariance matrix.
    #
    # This class represents the term :math:`\Sigma_t`, *i.e.,*
    # the :math:`(n-1) \times (n-1)` positive semi-definite matrix
    # which estimates the covariance of the (non-cash) assets' returns.
    # :ref:`Optimization-based policies` use this, as is explained
    # in Chapter 4 and 5 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_.
    #
    # The user can either supply a :class:`pandas.DataFrame` with the covariance matrix
    # (constant or varying in time) computed externally (for example
    # with some machine learning technique) or let this class estimate the covariance from the data.
    # The latter is the default behavior.
    #
    # This class implements three ways to compute the covariance matrix from the past returns. The
    # computation is repeated at each point in time :math:`t` of a :class:`BackTest` using only
    # the past returns available at that point: :math:`r_{t-1}, r_{t-2}, \ldots`.
    #
    # * *rolling covariance*, using :class:`pandas.DataFrame.rolling.cov`. This is done
    #   if the user specifies the ``rolling`` argument.
    # * *exponential moving window covariance*, using :class:`pandas.DataFrame.ewm.cov`. This is done
    #   if the user specifies the ``halflife`` argument (``rolling`` takes precedence).
    # * *full historical covariance*, using :class:`pandas.DataFrame.cov`. This is the default
    #   behavior if no arguments are specified.
    #
    # If there are missing data in the historical returns the estimated covariance may not
    # be positive semi-definite. We correct it by projecting on the positive semi-definite
    # cone (*i.e.*, we set the negative eigenvalues of the resulting :math:`\Sigma_t` to zero).
    #
    # :param Sigma: :class:`pandas.DataFrame` of covariance matrices
    #     supplied by the user. The DataFrame either represents a single (constant) covariance matrix
    #     or one for each point in time. In the latter case the DataFrame must have a :class:`pandas.MultiIndex`
    #     where the first level is a :class:`pandas.DatetimeIndex`. If ``None`` (the default)
    #     the covariance matrix is computed from past returns.
    # :type Sigma: pandas.DataFrame or None
    # :param rolling: if it is not ``None`` the covariance matrix will be estimated
    #     on a rolling window of size ``rolling`` of the past returns.
    # :type rolling: int or None
    # :param halflife: if it is not ``None`` the covariance matrix will be estimated
    #     on an exponential moving window of the past returns with half-life ``halflife``.
    #     If ``rolling`` is specified it takes precedence over ``halflife``. If both are ``None`` the full history
    #     will be used for estimation.
    # :type halflife: int or None
    # :param kappa: the multiplier for the associated forecast error risk
    #     (see pages 32-33 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_).
    #     If ``float`` a passed it is treated as a constant, if ``pandas.Series`` with ``pandas.DateTime`` index
    #     it varies in time, if ``None`` the forecast error risk term will not be compiled.
    # :type kappa: float or pandas.Series or None
    # :param kelly: correct the covariance matrix with the term :math:`\mu\mu^T`, as is explained
    #     in page 28 of the `book <https://web.stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_,
    #     to match the second term of the Taylor expansion of the portfolio log-return. Default
    #     is ``False``, corresponding to classical mean-variance optimization. If ``True``, it
    #     estimates :math:`\mu` with the same technique as :math:`\Sigma`, *i.e.*, with rolling window
    #     average, exponential moving window average, or an average of the full history.
    # :type kelly: bool
    # """

    def __init__(self, Sigma=None, kelly=True):

        if not Sigma is None:
            self.Sigma = DataEstimator(Sigma)
            self._alreadyfactorized = False
        else:
            self.Sigma = HistoricalFactorizedCovariance(kelly=kelly)
            self._alreadyfactorized = True

    def _pre_evaluation(self, universe, backtest_times):
        self.Sigma_sqrt = cp.Parameter((len(universe)-1, len(universe)-1))

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""

        if self._alreadyfactorized:
            self.Sigma_sqrt.value = self.Sigma.current_value
        else:
            Sigma = self.Sigma.current_value
            eigval, eigvec = np.linalg.eigh(Sigma)
            eigval = np.maximum(eigval, 0.)
            self.Sigma_sqrt.value = eigvec @ np.diag(np.sqrt(eigval))

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        self.cvxpy_expression = cp.sum_squares(
            self.Sigma_sqrt.T @ w_plus_minus_w_bm[:-1])
        return self.cvxpy_expression


class RiskForecastError(BaseRiskModel):
    """Risk forecast error. 

    Implements the model defined in page 31 of the book. Takes same arguments
    as :class:`DiagonalCovariance`.

    :param sigma_squares: per-stock variances, indexed by time if DataFrame.
        If None it will be fitted on past data.
    :type sigma_squares: pd.DataFrame or pd.Series or None
    """

    def __init__(self, sigma_squares=None):
        if sigma_squares is None:
            self.sigma_squares = HistoricalVariance(kelly=True)  # None None
        else:
            self.sigma_squares = DataEstimator(sigma_squares)
        # self.standard_deviations = ParameterEstimator(standard_deviations)
        # self.zeroforcash=True
        # self.kelly=True

    def _pre_evaluation(self, universe, backtest_times):
        self.sigmas_parameter = cp.Parameter(
            len(universe)-1, nonneg=True)  # +self.kelly))

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""

        # if self.sigma_squares is None:
        #     sigma_squares = past_returns.var(ddof=0)
        #     if self.kelly:
        #         mean = past_returns.mean()
        #         sigma_squares += mean**2
        #     if self.zeroforcash:
        #         sigma_squares.iloc[-1] = 0.
        #     sigma_squares = sigma_squares.values
        # else:
        #     sigma_squares = self.sigma_squares.current_value

        sigma_squares = self.sigma_squares.current_value

        self.sigmas_parameter.value = np.sqrt(sigma_squares)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):

        return cp.square(cp.abs(w_plus_minus_w_bm[:-1]).T @ self.sigmas_parameter)


class DiagonalCovariance(BaseRiskModel):
    """Diagonal covariance matrix, user-provided or fit from data.

    :param sigma_squares: per-stock variances, indexed by time if DataFrame.
        If None it will be fitted on past data.
    :type sigma_squares: pd.DataFrame or pd.Series or None 
    """

    def __init__(self, sigma_squares=None):
        if not sigma_squares is None:
            self.sigma_squares = DataEstimator(sigma_squares)
        else:
            self.sigma_squares = HistoricalVariance(kelly=True)  # None
        # self.zeroforcash = True
        # self.kelly = True
        # self.standard_deviations = ParameterEstimator(standard_deviations)

    def _pre_evaluation(self, universe, backtest_times):
        self.sigmas_parameter = cp.Parameter(len(universe)-1)  # +self.kelly))

    def _values_in_time(self, t, past_returns, **kwargs):
        """Update forecast error risk here, and take square root of Sigma."""
        # super()._recursive_values_in_time(t, current_weights, current_portfolio_value, past_returns, past_volumes, **kwargs)

        # if self.sigma_squares is None:
        #     sigma_squares = past_returns.var(ddof=0)
        #     if self.kelly:
        #         mean = past_returns.mean()
        #         sigma_squares += mean**2
        #     if self.zeroforcash:
        #         sigma_squares[-1] = 0.
        #     sigma_squares = sigma_squares.values
        # else:
        #     sigma_squares = self.sigma_squares.current_value

        sigma_squares = self.sigma_squares.current_value

        self.sigmas_parameter.value = np.sqrt(sigma_squares)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):

        return cp.sum_squares(cp.multiply(w_plus_minus_w_bm[:-1], self.sigmas_parameter))


class FactorModelCovariance(BaseRiskModel):
    """Factor model covariance, either user-provided or fitted from the data.

    It has the structure

    :math:`F \Sigma_{F} F^T + \mathbf{diag}(d)`

    where the factors exposure :math:`F` has as many rows as the number of assets and as many
    columns as the number of factors,
    the factors covariance matrix :math:`Sigma_{F}` is positive semi-definite,
    and the idyosyncratic variances vector :math:`d` is non-negative. 

    The advantage of this risk model over the standard :class:`FullCovariance` is mostly
    computational. When well-specified (as we do here) it costs much less to solve an 
    optimization problem with this model than with a full covariance. It is a standard 
    model that has been used for many decades in the portfolio optimization community
    and multiple vendors exist for covariance matrices in this form. We also provide
    the functionality to compute this automatically (which happens if you only specify
    the number of factors to the constructor) with a standard PCA of the historical 
    covariance at each point in time of the backtest (only looking at past returns).

    :param F: factors exposure matrix either constant or varying in time. If constant
        use a dataframe where the index are the factors and the columns are the asset names.
        If varying in time use a pandas multiindexed dataframe where the first index level
        is time and the second level are the factors. The columns
        should always be the asset names. If None the constructor will default to fit the model from
        past returns at each point of the backtest.
    :type F: pd.DataFrame or None
    :param Sigma_F: factors covariance matrix either constant or varying in time. If varying in time
        use a multiindexed dataframe where the first index level is time, and the second are the
        factors (like the columns). If None it is assumed that :math:`Sigma_F` is the identity matrix: Leaving
        this to None will not trigger automatic fit of the model. You can also have a factors
        exposure matrix that is fixed in time and a factors covariance that instead changes in time,
        or the opposite.
    :type Sigma_F: pd.DataFrame or None
    :param d: idyosyncratic variances either constant or varying in time. If constant use a pandas series,
        if varying in time use a pandas dataframe where the index is time and the columns are the asset
        names. You can have this varying in time and the exposures or the factors covariance fixed, or the
        opposite. If you leave this to None you will trigger automatic fit of the model. If you wish
        to have no idyosyncratic variances you can for example just pass 0.
    :type d: pd.Series or pd.DataFrame or None
    :param num_factors: number of factors (columns of F) that are obtained when fitting the model automatically
    :type num_factors: int    
    :param kelly: when fitting the model you can perform a PCA on the covariance defined in the standard way
        with this equal to False, or use the the quadratic term of the Taylor approximation of the Kelly
        gambling objective with this equal to True. The difference is very small and it's better both
        computationally and in terms of portfolio performance to leave this to True.
    :type kelly: bool
    """

    def __init__(self, F=None, d=None, Sigma_F=None, num_factors=1, kelly=True):
        self.F = F if F is None else DataEstimator(F, compile_parameter=True)
        self.d = d if d is None else DataEstimator(d)
        self.Sigma_F = Sigma_F if Sigma_F is None else DataEstimator(Sigma_F, ignore_shape_check=True)
        if (self.F is None) or (self.d is None):
            self._fit = True
            self.Sigma = HistoricalFactorizedCovariance(kelly=kelly)
        else:
            self._fit = False
        self.num_factors = num_factors

    def _pre_evaluation(self, universe, backtest_times):
        self.idyosync_sqrt_parameter = cp.Parameter(len(universe)-1)
        effective_num_factors = min(self.num_factors, len(universe)-1)
        if self._fit:
            self.F_parameter = cp.Parameter(
                (effective_num_factors, len(universe)-1))
        else:
            if self.Sigma_F is None:
                self.F_parameter = self.F.parameter
            else:
                # we could refactor the code here so we don't create duplicate parameters
                self.F_parameter = cp.Parameter(self.F.parameter.shape)

    def _values_in_time(self, t, past_returns, **kwargs):
        if self._fit:
            Sigmasqrt = self.Sigma.current_value
            # numpy eigendecomposition has largest eigenvalues last
            self.F_parameter.value = Sigmasqrt[:, -self.num_factors:].T
            d = np.sum(Sigmasqrt[:, :-self.num_factors]**2, axis=1)
        else:
            d = self.d.current_value
            if not (self.Sigma_F is None):
                self.F_parameter.value = (
                    self.F.parameter.value.T @ np.linalg.cholesky(self.Sigma_F.current_value)).T

        self.idyosync_sqrt_parameter.value = np.sqrt(d)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        self.expression = cp.sum_squares(cp.multiply(
            self.idyosync_sqrt_parameter, w_plus_minus_w_bm[:-1]))
        assert self.expression.is_dcp(dpp=True)

        self.expression += cp.sum_squares(self.F_parameter @
                                          w_plus_minus_w_bm[:-1])
        assert self.expression.is_dcp(dpp=True)

        return self.expression


class WorstCaseRisk(BaseRiskModel):
    """Select the most restrictive risk model for each value of the allocation vector.

    Given a list of risk models, penalize the portfolio allocation by the
    one with highest risk value at the solution point. If uncertain about
    which risk model to use this procedure can be an easy solution.

    :param riskmodels: risk model instances on which to compute the worst-case
        risk.
    :type riskmodels: list 
    """

    def __init__(self, riskmodels):
        self.riskmodels = riskmodels

    def _recursive_pre_evaluation(self, universe, backtest_times):
        """Initialize objects."""
        for risk in self.riskmodels:
            risk._recursive_pre_evaluation(universe, backtest_times)

    def _recursive_values_in_time(self, **kwargs):
        """Update parameters."""
        for risk in self.riskmodels:
            risk._recursive_values_in_time(**kwargs)

    def _compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        risks = [risk._compile_to_cvxpy(w_plus, z, w_plus_minus_w_bm)
                 for risk in self.riskmodels]
        return cp.max(cp.hstack(risks))
